import json, logging
from rdkit import Chem, Geometry
from rdkit.Chem import rdFMCS, AllChem  # noqa
from typing import Dict
from molecular_rectifier import Rectifier
from ..monster import Monster


class MCSMerger(Monster):
    """
    see ``combine()`` for ops.
    """
    journal = logging.getLogger()

    extreme_embed_args = [('maxAttempts', 0, 5),
                          ('numZeroFail', 1, 3),
                          ('forceTol', 0.001, 1),
                          ('ignoreSmoothingFailures', False, True),  # max([True, False]) == True
                          ]

    def __init__(self, hits: Chem.Mol, mcs_parameters=rdFMCS.MCSParameters()):
        self.hits = hits
        assert len(hits) == 2, 'Only two hits requirement, not more, not less'
        self.fore = hits[0]
        self.aft = hits[1]
        self.mcs_parameters = mcs_parameters
        self.common = self.get_common()
        self.aft2fore_mapping = self.get_aft2fore_mapping()
        # added in combine
        self.mod = Chem.RWMol
        self.flat_combo = Chem.Mol
        self.rectified = Chem.Mol
        self.positioned_mol = Chem.Mol
        # these are to match Monster
        self.minimized_mol = Chem.Mol
        self.mol_options = []
        self.unmatched = []

    def get_common(self) -> Chem.Mol:
        res = rdFMCS.FindMCS([self.fore, self.aft], self.mcs_parameters)
        return Chem.MolFromSmarts(res.smartsString)

    def get_aft2fore_mapping(self) -> Dict[int, int]:
        return dict(zip(self.aft.GetSubstructMatch(self.common), self.fore.GetSubstructMatch(self.common)))

    def combine(self, embed_mode: int = 0):
        """
        Combine fore and aft into a single molecule (see below).
        Will add property 'provenance' to atoms, with values 'fore' or 'aft' or 'common'.
        Will add property 'provenance' to bonds, with values 'fore', 'aft', 'common' or 'semicommon' (different bond order).
        This property is different from `_Origin` / `_Origins` and `_BondProvenance`...

        The argument ``embed_mode`` (int for now) controns if to use
        0 = ``AllChem.EmbedMolecule`` or
        1 = ``AllChem.ConstrainedEmbed``

        * ``.mod`` is the combined molecule (RWMol).
        * ``.flat_combo`` is ``.mod`` without Hs and coordinates.
        * ``.rectified`` is ``.flat_combo`` after rectification.
        * ``.positioned_mol`` is ``.rectified`` with Hs and embedded coordinates.
        """
        self.mark(self.fore, 'fore_')
        self.mod = Chem.RWMol(self.fore)
        # copy atoms and bonds from aft to fore-copy
        self._add_atoms()
        # copy bonds from aft to fore-copy
        self._add_bonds()
        # now mod > flat_combo > rectified > positioned_mol
        # clean up
        self.positioned_mol = self._clean_up()
        # embed
        if embed_mode == 0:
            self._embed()
        else:
            self._con_embed()
        # self._con_embed()
        self.positioned_mol.SetProp('_Name', self.merged_name)
        self.convert_to_origins(self.positioned_mol)
        return self.positioned_mol

    @property
    def merged_name(self):
        if not self.fore.HasProp('_Name') and not self.aft.HasProp('_Name'):
            return 'merger'
        if not self.fore.HasProp('_Name'):
            self.fore.SetProp('First')
        if not self.aft.HasProp('_Name'):
            self.aft.SetProp('Second')
        return self.fore.GetProp('_Name') + self.aft.GetProp('_Name')

    def mark(self, mol: Chem.Mol, prefix=''):
        """
        Adds the fore_i and aft_i properties to atoms.
        """
        for atom in mol.GetAtoms():
            atom.SetIntProp(prefix + 'i', atom.GetIdx())

    def _add_atoms(self):
        """
        Add atoms from aft to mod, a copy of fore.
        Called by combine().
        Will add property 'provenance' to atoms, with values 'fore', 'aft' or 'common'.
        If the atom is a product of rectification then it would be `rectified`.

        :return: None
        """
        atom: Chem.Atom
        # ### mark originals
        for atom in self.mod.GetAtoms():
            atom.SetProp('provenance', 'fore')
        # ### copy atoms
        self.mod.BeginBatchEdit()
        for atom in self.aft.GetAtoms():
            i = atom.GetIdx()
            if i in self.aft2fore_mapping:
                self.mod.GetAtomWithIdx(self.aft2fore_mapping[i]).SetProp('provenance', 'common')
                continue
            if atom.GetAtomicNum() == 1:
                # skip H
                # zahl of dummy is 0: these are left - wise?
                continue
            new_i = self.mod.AddAtom(Chem.Atom(atom.GetAtomicNum()))
            new_atom = self.mod.GetAtomWithIdx(new_i)
            new_atom.SetProp('provenance', 'aft')
            new_atom.SetIntProp('aft_i', i)
            self.aft2fore_mapping[i] = new_i
        self.mod.CommitBatchEdit()

    def _add_bonds(self):
        """
        Add bonds from aft to mod, a copy of fore.
        Called by combine().
        Will add property 'provenance' to bonds, with values 'fore' or 'aft' or 'common' or 'semicommon'.
        The latter for bonds that are common but with different bond order.
        If the bond is a product of rectification then it would be `rectified`.

        :return: None
        """
        bond: Chem.Bond
        for bond in self.mod.GetBonds():
            bond.SetProp('provenance', 'fore')
            bond.SetIntProp('fore_i', bond.GetIdx())
        # ### copy bonds
        for aft_bond in self.aft.GetBonds():
            aft_begin_idx = aft_bond.GetBeginAtomIdx()
            if self.aft.GetAtomWithIdx(aft_begin_idx).GetAtomicNum() == 1:
                # skip H
                continue
            aft_end_idx = aft_bond.GetEndAtomIdx()
            if self.aft.GetAtomWithIdx(aft_end_idx).GetAtomicNum() == 1:
                # skip H
                continue
            fore_begin_idx = self.aft2fore_mapping[aft_begin_idx]
            fore_end_idx = self.aft2fore_mapping[aft_end_idx]
            mod_bond = self.mod.GetBondBetweenAtoms(fore_begin_idx, fore_end_idx)  # noqa
            if mod_bond is None:
                # the bond did not exist
                total_bonds: int = self.mod.AddBond(fore_begin_idx, fore_end_idx, aft_bond.GetBondType())  # noqa
                mod_bond = self.mod.GetBondBetweenAtoms(fore_begin_idx, fore_end_idx)  # noqa
                mod_bond.SetProp('provenance', 'aft')
            elif aft_bond.GetBondTypeAsDouble() > mod_bond.GetBondTypeAsDouble():
                # the bond existed but wasn't as reduced
                mod_bond.SetBondType(aft_bond.GetBondType())
                mod_bond.SetProp('provenance', 'semicommon')
            else:
                mod_bond.SetProp('provenance', 'common')

    def _clean_up(self):
        """
        `.flat_combo`` is ``.mod`` withoout Hs and coordinates.
        `.rectified`` is ``.flat_combo`` after rectification.
        the returned will be `.positioned_mol`` ie. ``.rectified`` with Hs.
        This will gain coordinates in `.embed()`.

        :return:
        """
        combo = self.mod.GetMol()
        self.remove_chiral_tags(combo)
        self.flat_combo = AllChem.RemoveHs(combo, sanitize=False, updateExplicitCount=True)
        AllChem.Compute2DCoords(self.flat_combo)  # why is this needed? Is rectifier 3D independent?
        self.rectified = Rectifier(self.flat_combo).fix().mol
        # get only first largest fragment (in case it's split)
        largest_frag = sorted(AllChem.GetMolFrags(AllChem.RemoveHs(self.rectified), asMols=True),
                              key=lambda mol: mol.GetNumHeavyAtoms(),
                              reverse=True)[0]
        hydroed = AllChem.AddHs(largest_frag)
        AllChem.EmbedMolecule(hydroed)
        for atom in hydroed.GetAtoms():
            if not atom.HasProp('provenance'):
                atom.SetProp('provenance', 'rectified')
        for bond in hydroed.GetBonds():
            if not bond.HasProp('provenance'):
                bond.SetProp('provenance', 'rectified')
        hydroed.SetProp('_Name', self.merged_name)
        self.remove_chiral_tags(hydroed)
        return hydroed  # will be saved as self.positioned_mol

    def remove_chiral_tags(self, mol: Chem.Mol):
        for atom in mol.GetAtoms():
            atom.SetChiralTag(Chem.ChiralType.CHI_UNSPECIFIED)

    def is_valid(self, atom: Chem.Atom, trial: int):
        """
        Is this a constrainable atom?
        """
        if atom.GetAtomicNum() == 1:
            return False
        if not atom.HasProp('fore_i'):
            return False
        if trial < 1:
            return True
        for bond in atom.GetBonds():
            if bond.GetProp('provenance') == ('semicommon'):
                return False
        return trial == 1 or atom.GetProp('provenance') == 'common'

    def _embed(self, trial=0, **embed_args) -> None:
        """
        At trial 0, all fore atoms are okay
        At trial 1, only fore atoms with bonds with original bonding are okay
        At trial 2, only common atoms with bonds with original bonding are okay
        """
        self.journal.debug(f'Embedding trial {trial}')
        combo2fore = {atom.GetIdx(): atom.GetIntProp('fore_i') for atom in self.positioned_mol.GetAtoms() if
                      self.is_valid(atom, trial)}
        fore_conf = self.fore.GetConformer()
        coord_map: Dict[int, Geometry.Point3D] = {ci: fore_conf.GetAtomPosition(fi) for ci, fi in combo2fore.items()}
        conf_i = AllChem.EmbedMolecule(self.positioned_mol,
                                       **{**dict(clearConfs=True,
                                                 coordMap=coord_map,
                                                 ignoreSmoothingFailures=True),
                                          **embed_args
                                          }
                                       )

        if conf_i != -1:
            # add selection & return
            self.positioned_mol.__sssAtoms = list(coord_map)
            for atom in self.positioned_mol.GetAtoms():
                i: int = atom.GetIdx()
                atom.SetBoolProp('constrained', i in coord_map)
            self.positioned_mol.SetIntProp('embedding_trial', trial)
            # align
            rmsd: float = AllChem.AlignMol(self.positioned_mol,
                                           self.fore,
                                           atomMap=list(combo2fore.items())
                                           )
            # print('.... ', list(combo2fore.items()), rmsd)
            return
        # repeat...
        if trial == 2:
            # getting with rdkit default
            for key, default, new in self.extreme_embed_args:
                embed_args[key] = max(embed_args.get(key, default), new)
        if trial == 3:
            raise ValueError('Embedding failed')
        self._embed(trial + 1, **embed_args)

    def _con_embed(self, trial: int = 0) -> None:
        slimmed = Chem.RWMol(self.positioned_mol)
        keepers = []
        unkeepers = []
        for atom in slimmed.GetAtoms():
            if self.is_valid(atom, trial):
                keepers.append(atom.GetIdx())
            else:
                unkeepers.append(atom.GetIdx())

        slimmed.BeginBatchEdit()
        for i in unkeepers:
            slimmed.RemoveAtom(i)
        slimmed.CommitBatchEdit()
        fore_conf = self.fore.GetConformer()
        conf = Chem.Conformer(slimmed.GetNumAtoms())
        for atom in slimmed.GetAtoms():
            point: Geometry.Point3D = fore_conf.GetAtomPosition(atom.GetIntProp('fore_i'))
            conf.SetAtomPosition(atom.GetIdx(), point)
        slimmed.AddConformer(conf)
        try:
            AllChem.ConstrainedEmbed(self.positioned_mol, slimmed, useTethers=True, clearConfs=True)
            self.positioned_mol.__sssAtoms = keepers
            # align
            rmsd: float = AllChem.AlignMol(self.positioned_mol,
                                           self.fore,
                                           atomMap=[(ci, self.positioned_mol.GetAtomWithIdx(ci).GetIntProp('fore_i'),)
                                                    for ci in keepers]
                                           )
            for atom in self.positioned_mol.GetAtoms():
                i: int = atom.GetIdx()
                atom.SetBoolProp('constrained', i in keepers)
            self.positioned_mol.SetIntProp('embedding_trial', trial)
            return
        except ValueError:
            if trial == 2:
                raise ValueError('Embedding failed')
            self._con_embed(trial + 1)

    def convert_to_origins(self, mol: Chem.Mol, both=False):
        """
        In this class the 'provenance' are marked as common, aft, fore, rectified (semicommon is bond)
        """
        fore_name = self.fore.GetProp('_Name')
        aft_name = self.fore.GetProp('_Name')
        for atom in mol.GetAtoms():
            origins = []
            if atom.HasProp('fore_i'):
                origins.append(fore_name + '.' + atom.GetProp('fore_i'))
            if both and atom.HasProp('aft_i'):
                origins.append(aft_name + '.' + atom.GetProp('aft_i'))
            atom.SetProp('_Origin', json.dumps(origins))



#!/usr/bin/env python3

from ..version import __version__

__doc__ = """
Command line interface to Fragmenstein.
The the first argument to the command is one of the three option:

## extract
Given a target folder (`-i <folder_name>`) 
extract the mol files (with covalent atom if present) in an output folder (`-o <folder_name>`) 

## merge


## position

NB. Please no spaces in the filenames. If spaces are present quote/double-quote the fle path.
""".strip()

from .monster import FragmensteinParserMonster
from .parser import FragmensteinParser
from .base import FragmensteinParserBase


def main():
    FragmensteinParser.__doc__ = __doc__
    # init configures the parser, calling it executes it
    FragmensteinParser()()


if __name__ == '__main__':
    main()



"""
See documentation/notes/colours.md

``divergent_colors`` are R-style CIE HChL divergent.

These were generated via:

.. code-block:: python
     !pip install git+https://github.com/retostauffer/python-colorspace
     import json
     import numpy as np
     from colorspace.colorlib import HCL, hexcols, colorobject
     from typing import List

     def generate_series(base:colorobject, n:int) -> List[str]:
         base.to('HCL')
         hues : np.ndarray = np.linspace(0,360, n+1)+base.get('H')[0]
         hues[hues >= 360] -= 360

         colors = HCL(H = hues[:-1], C = [base.get('C')[0]]*n, L = [base.get('L')[0]]*n)
         colors.to('hex')
         return colors.colors()

     base = hexcols('#AED882')
     colors = {i: generate_series(base, i) for i in range(1, 101)}
     with open('divergent_colors.json', 'w') as fh:
         json.dump(colors, fh)
"""

__all__ = ['feijoa', 'divergent_colors']

feijoa = '#AED882'

import json
from typing import List, Dict
import pkg_resources

json_str = pkg_resources.resource_string(__name__, 'divergent_colors.json').decode('utf-8')

divergent_colors:Dict[int, List[str]] = {int(k): v for k, v in json.loads(json_str).items()}



########################################################################################################################
__doc__ = \
    """
This is contains the class _MonsterBase to be inherited by _MonsterCommunal, then Monster
    """

########################################################################################################################

import logging
from typing import List, Optional, Dict, Sequence, Tuple, Union
from rdkit import Chem
from rdkit.Chem import rdFMCS


class _MonsterBase:
    """
    _MonsterBase -> _MonsterTracker -> _MonsterCommunal
    """

    journal = logging.getLogger('Fragmenstein')

    # overriding these seems insane.
    dummy_symbol = '*'
    dummy = Chem.MolFromSmiles(dummy_symbol)  #: The virtual atom where the targets attaches

    # settings.py...
    atoms_in_bridge_cutoff = 2
    # atoms_in_bridge_cutoff is how many bridge atoms can be deleted?
    # (0 = preserves norbornane, 1 = preserves adamantane)
    throw_on_discard = False
    strict_matching_mode = dict(atomCompare=rdFMCS.AtomCompare.CompareElements,
                                 bondCompare=rdFMCS.BondCompare.CompareOrder,
                                 ringMatchesRingOnly=True,
                                 ringCompare=rdFMCS.RingCompare.StrictRingFusion,
                                 matchChiralTag=True)
    matching_modes = [
        dict(atomCompare=rdFMCS.AtomCompare.CompareAny,
             bondCompare=rdFMCS.BondCompare.CompareAny,
             ringCompare=rdFMCS.RingCompare.PermissiveRingFusion,
             ringMatchesRingOnly=False),  # this shape based matching is too permissive,
        dict(atomCompare=rdFMCS.AtomCompare.CompareAny,
             bondCompare=rdFMCS.BondCompare.CompareOrder,
             ringCompare=rdFMCS.RingCompare.PermissiveRingFusion,
             ringMatchesRingOnly=False),
        dict(atomCompare=rdFMCS.AtomCompare.CompareElements,
             bondCompare=rdFMCS.BondCompare.CompareOrder,
             ringCompare=rdFMCS.RingCompare.PermissiveRingFusion,
             ringMatchesRingOnly=False),
        dict(atomCompare=rdFMCS.AtomCompare.CompareAny,
             bondCompare=rdFMCS.BondCompare.CompareAny,
             ringCompare=rdFMCS.RingCompare.PermissiveRingFusion,
             ringMatchesRingOnly=True),
        dict(atomCompare=rdFMCS.AtomCompare.CompareAny,
             bondCompare=rdFMCS.BondCompare.CompareOrder,
             ringCompare=rdFMCS.RingCompare.PermissiveRingFusion,
             ringMatchesRingOnly=True),
        dict(atomCompare=rdFMCS.AtomCompare.CompareElements,
             bondCompare=rdFMCS.BondCompare.CompareOrder,
             ringCompare=rdFMCS.RingCompare.PermissiveRingFusion,
             ringMatchesRingOnly=True)]

    # ------------------------------------------------------------------------------------------------------------------

    def __init__(self,
                 hits: List[Chem.Mol],
                 average_position: bool=False,
                 joining_cutoff: float =5,
                 random_seed: Optional[int] = None):
        """
        Initialisation starts Monster, but it does not do any mergers or placements.
        This is changed in revision 0.6 (previously `mol` was specified for the latter)

        :param hits:
        :param average_position:
        :param joining_cutoff: joining cutoff used in "full" mode
        :param random_seed: A random seed for rdkit embedding calculations during placement

        """
        # ==== hits ===========================================
        # fix_hits: assert Chem.Mol, fix name if needed and store positions (see ``store_positions``)
        self.hits = self.fix_hits(hits)  # list of hits
        # ==== other ==========================================
        #self._debug_draw has been taken over by ``modifications`` and ``journal``
        self.average_position = average_position
        # ==== To do be filled ================================
        # -------- placement ----------------------------------
        self.initial_mol = None  # to be filled by place. The starting molecule (Chem.Mol).
        # Manually assignmnt of self.initial_mol is futile
        self.attachment = None  # place only.
        self.custom_map: Dict[str, Dict[int, int]] = {}
        # -------- common ------------------------------------
        # # ivars of type List[str]
        self.unmatched = []  # rejected hit names List[str]
        # self.matched is dynamic.  # accepted hits names List[str]
        # # ivars of type Chem.Mol or List[Chem.Mol] or Dict[Chem.Mol]
        self.modifications = {}
        self.positioned_mol = None  # final molecule
        self.joining_cutoff = joining_cutoff  # over-ridden
        self.random_seed = random_seed
        self.mol_options = []  # equally valid alternatives to self.positioned_mol
        self._collapsed_ring_offset = 0  # variable to keep track of how much to offset in ring collapse.
        # formerly:
        # self.scaffold = None  # template which may have wrong elements in place, or
        # self.chimera = None  # merger of hits but with atoms made to match the to-be-aligned mol

    def fix_hits(self, hits: List[Chem.Mol]) -> List[Chem.Mol]:
        """
        Adds the ``_Name`` Prop if needed
        asserts everything is a Chem.Mol
        calls ``store_positions``
        :param hits:
        :return:
        """
        dejavu = set()
        for hi, hit in enumerate(hits):
            if isinstance(hit, str):
                self.journal.warning(f'Hit {hi} is a string ({hit}).' +
                                     'This route is not the intended way. Trying to read it.')
                if '.mol' in hit or '.mdf' in hit:
                    hits[hi] = Chem.MolFromMolFile(hit)
                elif '.pdb' in hit:
                    hits[hi] = Chem.MolFromPDBFile(hit)
                else:
                    raise ValueError(f'Hit {hit} is not a Mol file.')
            elif isinstance(hit, Chem.Mol):
                pass
            else:
                raise ValueError(f'Hit has to be a Chem.Mol! not {type(hit)}')
            # fallback naming.
            if not hit.HasProp('_Name') or hit.GetProp('_Name').strip() == '':
                for k in ('id', 'ID', 'Id',  'name', 'Name'):
                    if hit.HasProp(k):
                        hit.SetProp('_Name', hit.GetProp(k))
                        break
                else:
                    hit.SetProp('_Name', f'hit{hi}')
            elif hit.GetProp('_Name') in dejavu:
                hit.SetProp('_Name', hit.GetProp('_Name') + f'_{hi}')
            dejavu.add(hit.GetProp('_Name'))
            if not hit.HasProp('Id'):
                hit.SetProp('Id', hit.GetProp('_Name'))
            # ====== IMPORTANT ==========
            self.store_positions(hit)
        return hits

    def store_positions(self, mol: Chem.Mol) -> Chem.Mol:
        """
        Saves positional data as _x, _y, _z and majorly ``_ori_i``, the original index.
        The latter gets used by ``_get_new_index``.

        :param mol:
        :return:
        """
        conf = mol.GetConformer()
        name = mol.GetProp('_Name')
        for i, atom in enumerate(mol.GetAtoms()):
            pos = conf.GetAtomPosition(i)
            atom.SetIntProp('_ori_i', i)
            atom.SetProp('_ori_name', name)
            atom.SetDoubleProp('_x', pos.x)
            atom.SetDoubleProp('_y', pos.y)
            atom.SetDoubleProp('_z', pos.z)
        return mol

    @property
    def matched(self) -> List[str]:
        """
        This is the counter to unmatched.
        It's dynamic as you never know...

        :return:
        """
        return [h.GetProp('_Name') for h in self.hits if
                h.GetProp('_Name') not in self.unmatched]

    def get_hit_by_name(self, name: str) -> Chem.Mol:
        """
        Given a name of a hit (as defined in ``_Name`` property), return the hit.
        Do note `fix_hits` will have been called, so the name may be assigned.
        :param name:
        :return:
        """
        for hit in self.hits:
            if hit.GetProp('_Name') == name:
                return hit
        else:
            raise ValueError(f"No hit with name {name}")
        

    def fix_custom_map(self,
                          custom_map: Dict[str, Union[Sequence[Tuple[int, int]], Dict[int, int]]]) \
            -> Dict[str, Dict[int, int]]:
        """
        This is duplicated in SpecialCompareAtoms,
        but will be deprecated in favour of this one.

        Make sure its Dict[str, Dict[int, int]]

        There is a bit of confusion about the custom map.
        Converts the custom map from dict of lists of 2-element tuples to dict of dicts.
        """
        if custom_map is None:
            # in Monster only not SpecalCompareAtoms!
            return {h.GetProp('_Name'): {} for h in self.hits}
        assert isinstance(custom_map, dict), 'User defined map has to be mol name to Dict[int, int]'
        for name, hit_map in custom_map.items():
            custom_map[name] = dict(hit_map)
        return custom_map




from rdkit import Chem
from rdkit.Chem import rdFMCS
# TypedDict & Unpack fixed in .legacy:
from typing import Dict, List, Tuple, Optional, TypeVar, Sequence, TypedDict, Unpack  # noqa

IndexMap = TypeVar('IndexMap', bound=Sequence[Tuple[int, int]])
# this is not visible via help as IndexMap is technically an instance
IndexMap.__doc__ = """
Sequence (Tuple or List) of tuples of two indices, the first is the hit index, the second is the followup
"""


class BasicFMCSMode(TypedDict):
    """
    These are the acceptable types for the modes of FindMCS
    """
    atomCompare: rdFMCS.AtomCompare
    bondCompare: rdFMCS.BondCompare
    ringCompare: rdFMCS.RingCompare
    matchChiralTag: bool
    matchValences: bool
    completeRingsOnly: bool
    ringMatchesRingOnly: bool
    threshold: float
    maximizeBonds: bool
    timeout: int
    verbose: bool
    seedSmarts: str


class ExtendedFMCSMode(BasicFMCSMode):
    """
    These are the acceptable types for the modes of FindMCS, plus three that can be configured in rdFMCS.MCSParameters
    """
    matchStereo: bool
    matchFormalCharge: bool
    maxDistance: float



import itertools
from typing import Union, Sequence, List, Iterator

import pandas as pd
import pebble
from rdkit import Chem, rdBase

from ._base import LabBench, binarize, unbinarize
from ..igor import pyrosetta  # this may be pyrosetta or a mock for Sphinx in RTD


class LabCombine(LabBench):

    def combine_subprocess(self, binary_hits: List[bytes]):
        """
        This is the combination subprocess. The placement subprocess is ``place_subprocess``.
        They are very similar...
        """
        if self.Victor.uses_pyrosetta:
            pyrosetta.distributed.maybe_init(extra_options=self.init_options)
        tentative_name = 'UNKNOWN'
        try:
            hits: List[Chem.Mol] = [hit for hit in map(unbinarize, binary_hits) if hit]
            assert len(hits) > 0, 'No valid hits!'
            tentative_name = '-'.join([mol.GetProp('_Name') for mol in hits])
            if tentative_name in self.blacklist:
                raise ValueError(f'{tentative_name} is blacklisted')
            # `self.Victor` is likely `Victor` but the user may have switched for a subclass, cf. `VictorMock`...
            victor = self.Victor(hits=hits,
                                 pdb_block=self.pdbblock,
                                 ligand_resn='LIG',
                                 ligand_resi=self.ligand_resi,
                                 covalent_resi=self.covalent_resi,
                                 # a random residue is **still** required for the constaint ref atom.
                                 )
            victor.monster_throw_on_discard = True
            victor.monster.throw_on_discard = True
            victor.combine()
            result: dict = victor.summarize()
            result['unmin_binary'] = binarize(victor.monster.positioned_mol)
            result['min_binary'] = binarize(victor.minimized_mol)
            result['hit_binaries'] = [binarize(h) for h in victor.hits]
            if self.run_plip:
                result.update(victor.get_plip_interactions())
            return result
            # victor.make_pse()
        except KeyboardInterrupt as err:
            raise err
        except Exception as error:
            error_msg = f'{error.__class__.__name__} {error}'
            self.Victor.journal.critical(f'*** {error_msg} for {tentative_name}')
            return dict(error=error_msg, name=tentative_name)

    def combine(self,
                mols: Sequence[Chem.Mol],
                permute: bool = True,
                combination_size: int = 2,
                **kwargs) -> Union[pebble.ProcessMapFuture, pd.DataFrame]:
        """
        Combine all of ``mols`` with each other in combinations of ``combination_size``.
        Due to the way Monster works merging A with B may yield a different result to B with A.
        Hence the ``permute`` boolean argument.
        """  # extended at end of file.
        iterator: Iterator
        if permute:
            iterator = itertools.permutations(map(binarize, mols), combination_size)
        else:
            iterator = itertools.combinations(map(binarize, mols), combination_size)
        df = self(iterator=iterator, fun=self.combine_subprocess, **kwargs)
        df['outcome'] = df.apply(self.categorize, axis=1)
        with rdBase.BlockLogs():
            df['simple_smiles'] = df.unminimized_mol.apply(self.Victor.to_simple_smiles)
        self.fix_intxns(df)
        return df

    def twoway_combine(self,
                       primary_mols: Sequence[Chem.Mol],
                       secondary_mols: Sequence[Chem.Mol],
                       combination_size: int = 2,
                       **kwargs) -> Union[pebble.ProcessMapFuture, pd.DataFrame]:
        """
        Combine ``primary_mols`` with ``secondary_mols``.
        """
        iterator: Iterator
        if combination_size == 2:
            iterator = itertools.product(map(binarize, primary_mols), map(binarize, secondary_mols))
        elif combination_size > 2:
            extras = map(binarize, list(primary_mols) + list(secondary_mols))
            iterator = itertools.product(map(binarize, primary_mols), map(binarize, secondary_mols), extras,
                                         repeat=combination_size - 2)
        else:
            raise ValueError(f'combination_size must be > 2 (given: {combination_size}')
        df = self(iterator=iterator, fun=self.combine_subprocess, **kwargs)
        df['outcome'] = df.apply(self.categorize, axis=1)
        return df


# prepend docstring to combine
LabCombine.combine.__doc__ = LabBench.__call__.__doc__ + '\n\n' + LabCombine.combine.__doc__
LabCombine.twoway_combine.__doc__ = LabBench.__call__.__doc__ + '\n\n' + LabCombine.twoway_combine.__doc__



########################################################################################################################
__doc__ = \
    """
This is Monster proper. and contains the class ``Monster``.
The inheritance is as follows:


**_MonsterCombine**  inherits:

* ``_MonsterMerge``  (``_MonsterCommunal`` <- ``_MonsterTracker`` <- ``_MonsterBase``)
* ``_MonsterRing``

**_MonsterPlace**  inherits:

* ``_MonsterMerge``  (``_MonsterBlend`` < - ``_MonsterCommunal`` <- ``_MonsterTracker`` <- ``_MonsterBase``)

**_MonsterUtils** inherits

* ``_MonsterCommunal``  ( <- ``_MonsterTracker`` <- ``_MonsterBase``) 
* ``GPM``

Where:

**_MonsterBase** adds the cvars and the ``__init__`` and its dependent methods

**_MonsterTracker** inherits ``_MonsterBase`` and adds just a method to better store modifications

**_MonsterCommunal** inherits ``_MonsterTracker`` ( <- ``_MonsterBase``)


It adds methods for misc purposes beyond place/combine.
It is inherited by ``Monster`` only.
    """

########################################################################################################################

# --- independent of Monster -------------------------------------------------------------------------------------------
from .positional_mapping import GPM
from .bond_provenance import BondProvenance
from .unmerge_mapper import Unmerge
from .mcs_mapping import IndexMap, SpecialCompareAtoms, ExtendedFMCSMode, transmute_FindMCS_parameters


# ---  Monster ---------------------------------------------------------------------------------------------------------

from ._communal import _MonsterCommunal
from ._ff import _MonsterFF  # _MonsterFF extends _MonsterUtil
from ._combine import _MonsterCombine  # inherits _MonsterCommunal adds the combine method
from ._place import _MonsterPlace  # inherits _MonsterCommunal adds the combine method


class Monster(_MonsterFF, _MonsterPlace, _MonsterCombine):
    """
    This creates a stitched together monster.
    For initilialisation for either placing or combining, it needs a list of hits (rdkit.Chem.Mol).

    Note, the hits have to be 3D embedded as present in the protein —it would defeat the point otherwise!
    For a helper method to extract them from crystal structures see `Victor.extract_mol`.

    The calculation are done either by place or merge.

    ## Place

    >>> monster.place(mol)

    Given a RDKit molecule and a series of hits it makes a spatially stitched together version
    of the initial molecule based on the hits.
    The reason is to do place the followup compound to the hits as faithfully as possible
    regardless of the screaming forcefields.

    * ``.mol_options`` are the possible equiprobable alternatives.
    * ``.positioned_mol`` is the desired output (rdkit.Chem.Mol object)
    * ``.initial_mol`` is the input (rdkit.Chem.Mol object), this is `None` in a `.combine` call.
    * ``.modifications['scaffold']`` is the combined version of the hits (rdkit.Chem.Mol object).
    * ``.modifications['chimera']`` is the combined version of the hits, but with differing atoms made \
                    to match the followup (rdkit.Chem.Mol object).

    ``.get_positional_mapping``, which works also as a class method,
    creates a dictionary of mol_A atom index to mol_B atom index
    based on distance (cutoff 2&Aring;) and not MCS.

    The code works in two broad steps, first a scaffold is made, which is the combination of the hits (by position).
    Then the followup is placed. It is not embedded with constrained embedding functionality of RDKit as this
    requires the reference molecule to have a valid geometry, which these absolutely do not have this.
    Novel side chains are added by aligning an optimised conformer against the closest 3-4 reference atoms.
    Note that ``.initial_mol`` is not touched. ``.positioned_mol`` may have lost some custom properties,
    but the atom indices are the same.

    If an atom in a Chem.Mol object is provided via ``attachment`` argument and the molecule contains a dummy atom.
    Namely element R in mol file or * in string.

    ## Combine

    >>> monster.combine(keep_all=True, collapse_rings=True, joining_cutoff= 5))

    Combines the hits by merging and linking. ``collapse_rings`` argument results in rings being collapsed
    before merging to avoid oddities.
    The last step within the call is fixing any oddities of impossible chemistry via the call ``rectify``.
    This uses the separate class ``Rectifier`` to fix it.

    ## Attributes

    Common input derived

    :ivar hits:
    :vartype hits: list
    :cvar throw_on_discard: filled by keep_all
    :vartype throw_on_discard: bool

    Common derived

    :var matched: (dynamic) accepted hit names
    :vartype matched: List[str]
    :ivar unmatched: discarded hit names
    :vartype unmatched: List[str]

    :cvar journal: The "journal" is the log of Dr Victor Frankenstein (see Victor for more)
    :vartype journal: Logger
    :ivar modifications: copies of the mols along the way
    :vartype modifications: dict
    :ivar mol_options: equally valid alternatives to self.positioned_mol
    :vartype mol_options: list


    ``place`` specific:

    :ivar positioned_mol:
    :vartype positioned_mol: Mol
    :ivar attachment:
    :vartype attachment: NoneType
    :ivar initial_mol:
    :vartype initial_mol: NoneType
    :ivar average_position:
    :vartype average_position: bool
    :var num_common: (dynamic) number of atoms in common between follow-up and hits
    :vartype num_common: int
    :var percent_common: (dynamic) percentage of atoms of follow-up that are present in the hits
    :vartype percent_common: float

    ``combine`` specific:

    :cvar joining_cutoff: how distant (in Å) is too much?
    :vartype joining_cutoff: int
    :cvar atoms_in_bridge_cutoff: how many bridge atoms can be deleted?
        (0 = preserves norbornane, 1 = preserves adamantane)
    :vartype atoms_in_bridge_cutoff: int

    Class attributes best ignored:

    :cvar closeness_weights: list of functions to penalise closeness (ignore for most applications)
    :vartype closeness_weights: list
    :cvar dummy: The virtual atom where the targets attaches. by default `*`. Best not override.
    :vartype dummy: Mol
    :cvar dummy_symbol: The virtual atom where the targets attaches. by default `*`. Best not override.
    :vartype dummy_symbol: str
    :cvar matching_modes:
    :vartype matching_modes: list
    """
    pass



import argparse
from rdkit import Chem
from ..victor import Victor
from ..igor import Igor
from .base import FragmensteinParserBase, _add_common, set_verbose

class FragmensteinParserVictor:

    def _define_victor(self, parser: argparse.ArgumentParser):
        """fragmenstein victor combine -h hit1.mol hit2.mol -t protein.pdb
           fragmenstein victor place -h hit1.mol hit2.mol -s 'CCO'"""
        subparsers = parser.add_subparsers(title='operation', help='combine or place')
        combine_parser = subparsers.add_parser('combine', help='combine')
        _add_common(combine_parser, hits=True, output=True, name=False, template=True)
        combine_parser.set_defaults(func=self.victor_combine)
        place_parser = subparsers.add_parser('place', help='place')
        place_parser.add_argument('-s', '--smiles', required=True)
        _add_common(place_parser, hits=True, output=True, name=True, template=True)
        place_parser.set_defaults(func=self.victor_place)

    def victor_combine(self, args: argparse.Namespace) -> str:
        set_verbose(args.verbose)
        Igor.init_pyrosetta()
        Victor.work_path = args.output
        victor = Victor(hits=list(map(Chem.MolFromMolFile, args.hits)), pdb_filename=args.template)
        victor.combine(long_name=args.name)
        return Chem.MolToMolBlock(victor.minimized_mol)

    def victor_place(self, args: argparse.Namespace) -> str:
        set_verbose(args.verbose)
        Igor.init_pyrosetta()
        Victor.work_path = args.output
        victor = Victor(hits=list(map(Chem.MolFromMolFile, args.hits)), pdb_filename=args.template)
        victor.place(args.smiles, long_name=args.name)
        return Chem.MolToMolBlock(victor.minimized_mol)



# TypedDict & Unpack fixed in .legacy:
from typing import Dict, List, Tuple, Optional, TypeVar, Sequence, TypedDict, Unpack, overload  # noqa
from functools import singledispatch
from .types import IndexMap, BasicFMCSMode, ExtendedFMCSMode
from rdkit.Chem import rdFMCS



@overload
def flip_mapping(mapping: Sequence[Tuple[int, int]]) -> Sequence[Tuple[int, int]]:
    # basic IndexMap
    ...


@overload
def flip_mapping(mapping: Sequence[Sequence[Tuple[int, int]]]) -> Sequence[Sequence[Tuple[int, int]]]:
    # list of IndexMap options
    ...

@overload
def flip_mapping(mapping: Dict[int, int]) -> Dict[int, int]:
    ...

@overload
def flip_mapping(mapping: Dict[str, Dict[int, int]]) -> Dict[str, Dict[int, int]]:
    ...

@overload
def flip_mapping(mapping: Dict[str, Sequence[Tuple[int, int]]]) -> Dict[str, Sequence[Tuple[int, int]]]:
    ...

def flip_mapping(mapping):
    """
    This is a temporary fix for the fact that many route run on mappings that go hit to followup or viceversa
    Moving forward these will be replaced with a single hit to followup scheme, of type IndexMap
    """
    get_first_key = lambda d: list(d.keys())[0] if len(d) > 0 else None
    # empty
    if len(mapping) == 0:
        return mapping
    # Sequence[Tuple[int, int]] (basic)
    elif isinstance(mapping, Sequence) and isinstance(mapping[0], tuple) and isinstance(mapping[0][1], int):
        return [(b, a) for a, b in mapping]
    # Sequence[Sequence[Tuple[int, int]]]  (options
    elif isinstance(mapping, Sequence) and \
            isinstance(mapping[0], Sequence):
        return [flip_mapping(submapping) for submapping in mapping]
    elif isinstance(mapping, Sequence):  # boh
        raise NotImplementedError(f'{mapping}: list of ???')
    # Dict[int, int]
    elif isinstance(mapping, dict) and isinstance(get_first_key(mapping), int):
        return {b: a for a, b in mapping.items()}
    # Dict[str, Dict[int, int]]
    elif isinstance(mapping, dict) and isinstance(get_first_key(mapping), str):
        return {name: flip_mapping(submapping) for name, submapping in mapping.items()}
    # Dict[str, Sequence[Tuple[int, int]]
    elif isinstance(mapping, dict) and isinstance(get_first_key(mapping), str):
        return {name: flip_mapping(submapping) for name, submapping in mapping.items()}
    else:
        raise NotImplementedError(f'{mapping}: dictionary of {get_first_key(mapping)}')

# ----------------

def transmute_FindMCS_parameters(
        **mode: Unpack[ExtendedFMCSMode]) -> rdFMCS.MCSParameters:  # noqa lowercase not applicable
    """
    The function ``rdFMCS.FindMCS`` has two ways of being used.
    In one, a series of arguments are passed,
    in another a ``rdFMCS.MCSParameters`` object is passed (a wrapped C++ structure).
    Unfortunately, there does not seem to be a way to transmute the former into the other.

    Hence, this function

    The ``params.AtomTyper`` and ``params.BondTyper`` members
    can be either

    * the enum ``rdFMCS.AtomCompare`` or ``rdFMCS.BondCompare``
    * or a subclass of ``rdFMCS.MCSBondCompare`` or ``rdFMCS.MCSAtomCompare``

    The ``rdFMCS.RingCompare`` appears to be absorbed into the ``params.BondCompareParameters``
    member.
    """
    params = rdFMCS.MCSParameters()
    # three integers: https://github.com/rdkit/rdkit/blob/b208da471f8edc88e07c77ed7d7868649ac75100/Code/GraphMol/FMCS/FMCS.h
    # they are not rdFMCS.AtomCompare rdFMCS.BondCompare and rdFMCS.RingCompare enums?
    # atom parameters
    atomCompare: rdFMCS.AtomCompare = mode.get('atomCompare', rdFMCS.AtomCompare.CompareElements)
    params.AtomTyper = atomCompare  # int or a callable
    params.AtomCompareParameters.MatchIsotope = atomCompare == rdFMCS.AtomCompare.CompareIsotopes
    params.AtomCompareParameters.CompleteRingsOnly = mode.get('completeRingsOnly', False)
    params.AtomCompareParameters.MatchChiralTag = mode.get('matchChiralTag', False)
    params.AtomCompareParameters.MatchValences = mode.get('matchValences', False)
    params.AtomCompareParameters.RingMatchesRingOnly = mode.get('ringMatchesRingOnly', False)
    # bond parameters
    bondCompare: rdFMCS.BondCompare = mode.get('bondCompare', rdFMCS.BondCompare.CompareOrder)
    ringCompare: rdFMCS.RingCompare = mode.get('ringCompare', rdFMCS.RingCompare.IgnoreRingFusion)
    params.BondTyper = bondCompare
    params.BondCompareParameters.CompleteRingsOnly = mode.get('completeRingsOnly', False)
    params.BondCompareParameters.MatchFusedRings = ringCompare != rdFMCS.RingCompare.IgnoreRingFusion
    params.BondCompareParameters.MatchFusedRingsStrict = ringCompare == rdFMCS.RingCompare.StrictRingFusion
    params.BondCompareParameters.RingMatchesRingOnly = mode.get('ringMatchesRingOnly', False)
    params.Threshold = mode.get('threshold', 1.0)
    params.MaximizeBonds = mode.get('maximizeBonds', True)
    params.Timeout = mode.get('timeout', 3600)
    params.Verbose = mode.get('verbose', False)
    params.InitialSeed = mode.get('seedSmarts', '')
    # parameters with no equivalence (i.e. made up)
    params.BondCompareParameters.MatchStereo = mode.get('matchStereo', False)
    params.AtomCompareParameters.MatchFormalCharge = mode.get('matchFormalCharge', False)
    params.AtomCompareParameters.MaxDistance = mode.get('maxDistance', -1)
    # params.ProgressCallback Depracated
    return params



## This is an optional import

"""
Copy of
https://github.com/matteoferla/PLIP-PyRosetta-hotspots-test/blob/main/plipspots_docking/plipspots/serial.py

This is a class I use to _apply_ PLIP to a pd.Series of molecules.
It is not built for the project, but works.
Note I have not dumped the methods that are not needed for the project.
"""

import os
from pathlib import Path
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem import Draw
from rdkit.Chem import PandasTools

from functools import singledispatchmethod
from typing import Tuple, Dict, List, Union, Sequence, Optional, Any
from collections import Counter, defaultdict
from plip.structure.preparation import PDBComplex, PLInteraction
from openbabel.pybel import Atom, Residue
from openbabel.pybel import ob
from .minimalPDB import MinimalPDBParser
import warnings


class SerialPLIPper:
    """
    Calling the instance will return a ``Dict[Tuple[str, str, int], int]``,
    where the key is interaction type, residue 3-letter name, residue index
    and the value is the count of interactions.
    Basically, applying Plip to a pd.Series of Chem.Mol.

    Unplacking it is kind of wierd, the best way I reckon is a brutal for-loop:

    .. code-block:: python

        import pandas as pd
        import pandera.typing as pdt

        intxndexes: pdt.Series[Dict[Tuple[str, str, int], int]] = hits.ROMol.apply(SerialPLIPper(pdb_filename))
        # columns will still be a tuple...:
        intxn_df = pd.DataFrame(intxndexes.to_list()).fillna(0).astype(int)
        hits['N_interactions'] = intxn_df.sum(axis='columns')
        for c in sorted(intxn_df.columns, key=lambda kv: kv[2]):
            # columns will be a colon-separated string:
            hits[':'.join(map(str, c))] = intxn_df[c]
    """

    def __init__(self, pdb_block: str, resn='LIG', resi=1, chain='B'):
        assert 'ATOM' in pdb_block, f'No ATOM entry in block provided: {pdb_block}'
        self.pdb_block = pdb_block
        self.resn = resn
        self.resi = int(resi)  # dont give me alt codes
        self.chain = chain

    @classmethod
    def from_filename(cls, pdb_filename: str, *args, **kwargs):
        """
        The main constructor is from PDB block, this is from PDB file
        """
        with open(pdb_filename, 'r') as f:
            pdb_block = f.read()
        return cls(pdb_block, *args, **kwargs)

    def __call__(self, mol) -> Dict[Tuple[str, str, int], int]:
        if mol is None or not isinstance(mol, Chem.Mol) or mol.GetNumAtoms() == 0:
            return {}
        holo: str = self.plonk(mol)
        interaction_set: PLInteraction = self.get_interaction_set(holo)
        return self.get_interaction_counts(interaction_set)

    def assign_pdb(self, mol: Chem.Mol):
        """
        Fix the PDB info for the molecule, in place
        """
        counts = defaultdict(int)
        atom: Chem.Atom
        for atom in mol.GetAtoms():
            element: str = atom.GetSymbol()
            counts[element] += 1
            info = Chem.AtomPDBResidueInfo(atomName=f'{element: >2}{counts[element]: <2}',
                                           residueName=self.resn,
                                           residueNumber=self.resi, chainId=self.chain)
            atom.SetPDBResidueInfo(info)

    def plonk(self, mol):
        """
        Temporarily here. Do not copy.
        There likely is a way to do this in OBabel
        This is using Fragmenstein ``MinimalPDBParser``.

        :param mol:
        :return:
        """
        pdbdata = MinimalPDBParser(self.pdb_block, remove_other_hetatms=True, ligname=self.resn)
        self.assign_pdb(mol)
        moldata = MinimalPDBParser(Chem.MolToPDBBlock(mol))
        pdbdata.append(moldata)
        return str(pdbdata)

    @singledispatchmethod
    def get_interaction_set(self) -> PLInteraction:
        """
        Overloaded method: block or mol return the iternaction set
        :return:
        """
        raise NotImplementedError

    @get_interaction_set.register
    def _(self, block: str) -> PLInteraction:
        holo = PDBComplex()
        holo.load_pdb(block, as_string=True)
        holo.analyze()
        return holo.interaction_sets[':'.join([self.resn, self.chain, str(self.resi)])]

    @get_interaction_set.register
    def _(self, mol: Chem.Mol) -> PLInteraction:
        if mol.GetNumAtoms() == 0:
            raise ValueError('Molecule has no atoms')
        holo = PDBComplex()
        holo.load_pdb(self.plonk(mol), as_string=True)
        holo.analyze()
        return holo.interaction_sets[':'.join([self.resn, self.chain, str(self.resi)])]

    def get_atomname(self, atom: Union[Atom, ob.OBAtom], atomnames: Optional[Sequence[str]]=None) -> str:
        """
        Given an atom, return its name.
        """
        if atomnames is not None and isinstance(atom, Atom):
            return atomnames[atom.idx - 1]  # Fortran indexing
        elif isinstance(atom, Atom):
            res: ob.OBResidue = atom.residue.OBResidue
            obatom = atom.OBAtom
        elif isinstance(atom, ob.OBAtom):
            obatom: ob.OBAtom = atom
            res: ob.OBResidue = obatom.GetResidue()
        else:
            raise TypeError
        return res.GetAtomID(obatom)  # this is likely to be ' C  ' as Babel has stripped this info

    def get_atom_by_atomname(self,
                             residue: Union[ob.OBResidue, Residue],
                             atomname: str,
                             atomnames: Optional[Sequence[str]]=None) -> ob.OBAtom:
        """
        Get an atom by its name in a residue.
        Note that the ligand will have its original atom names stripped by Babel.
        """
        if isinstance(residue, Residue):
            residue = residue.OBResidue
        obatom: ob.OBAtom
        for obatom in ob.OBResidueAtomIter(residue):
            if self.get_atomname(obatom, atomnames).strip() == atomname.strip():
                return obatom
        else:
            raise ValueError(f'No atom with name {atomname} in residue {residue.GetName()}')

    def get_interaction_counts(self, interaction_set: PLInteraction) -> Dict[Tuple[str, str, int], int]:
        """
        Count the number of interactions of each type for each residue
        """
        intxns: List = interaction_set.all_itypes
        intxn_dex = defaultdict(int)
        for intxn in intxns:
            key = (intxn.__class__.__name__, intxn.restype, intxn.resnr)
            intxn_dex[key] += 1  # noqa default dict works with tuples
        return dict(sorted(intxn_dex.items(), key=lambda kv: kv[0][2]))

    def summarize_interactions(self, atom_names: Sequence[str]) -> List[Dict[str, Any]]:
        interaction_set = self.get_interaction_set(self.pdb_block)
        details: List[Dict[str, Any]] = []
        for intxn in interaction_set.all_itypes:
            details.append(self.summarize_interaction(intxn))
        return details

    def summarize_interaction(self, intxn, atom_names: Sequence[str]) -> Dict[str, Any]:
        # https://github.com/openbabel/openbabel/blob/master/data/atomtyp.txt
        # https://blog.matteoferla.com/2023/07/a-note-no-plip-interactions.html
        relevant_atom_names = []
        type_name = intxn.__class__.__name__
        details = {'type': type_name, 'protein_resn': intxn.restype, 'protein_resi': intxn.resnr,
                   'protein_chain': intxn.reschain}
        if type_name == 'hbond':
            if intxn.protisdon:
                # assert hbond.a.residue.name != self.resn
                details['atom_names'] = [atom_names[intxn.a.idx - 1]]
                details['type'] = 'hbond_acceptor'
                details['babel_atom_types'] = [intxn.atype]
            else:
                details['atom_names'] = [atom_names[intxn.d.idx - 1]]
                details['type'] = 'hbond_donor'
                details['babel_atom_types'] = [intxn.dtype]
            details['distance'] = intxn.distance_ad  # intxn.distance_ad is to donor, _ah to hydrogen
        elif type_name == 'hydroph_interaction':
            details['atom_names'] = [atom_names[intxn.ligatom.idx - 1]]
            details['babel_atom_types'] = [intxn.ligatom.type]
            details['distance'] = intxn.distance
        elif type_name == 'pistack':
            details['atom_names'] = [atom_names[a.idx - 1] for a in intxn.ligandring.atoms]
            details['babel_atom_types'] = [a.type for a in intxn.ligandring.atoms]
            details['distance'] = intxn.distance
        elif type_name == 'waterbridge':
            if intxn.protisdon:
                # assert hbond.a.residue.name != self.resn
                details['atom_names'] = [atom_names[intxn.a.idx - 1]]
                details['type'] = 'water_acceptor'
                details['babel_atom_types'] = [intxn.atype]
                details['distance'] = intxn.distance_aw
            else:
                details['atom_names'] = [atom_names[intxn.d.idx - 1]]
                details['type'] = 'water_donor'
                details['babel_atom_types'] = [intxn.dtype]
                details['distance'] = intxn.distance_dw
        elif type_name == 'saltbridge':
            if intxn.protispos:
                details['atom_names'] = [atom_names[a.idx - 1] for a in intxn.negative.atoms]
                details['type'] = 'saltbridge_negative'
                details['babel_atom_types'] = [a.type for a in intxn.negative.atoms]
                details['distance'] = intxn.distance
            else:
                details['atom_names'] = [atom_names[a.idx - 1] for a in intxn.positive.atoms]
                details['type'] = 'saltbridge_positive'
                details['babel_atom_types'] = [a.type for a in intxn.positive.atoms]
                details['distance'] = intxn.distance
        elif type_name == 'pication':
            if intxn.protcharged:
                details['atom_names'] = [atom_names[a.idx - 1] for a in intxn.ring.atoms]
                details['type'] = 'pication_ring'
                details['babel_atom_types'] = [a.type for a in intxn.ring.atoms]
                details['distance'] = intxn.distance
            else:
                details['atom_names'] = [atom_names[a.idx - 1] for a in intxn.charge.atoms]
                details['type'] = 'pication_charge'
                details['babel_atom_types'] = [a.type for a in intxn.charge.atoms]
                details['distance'] = intxn.distance
        elif type_name == 'halogenbond':
            details['atom_names'] = [atom_names[intxn.don.idx - 1]]
            details['babel_atom_types'] = [intxn.don.type]
            details['distance'] = intxn.distance
        else:  # metal_complex basically.
            raise TypeError(type_name)
        return details


from ._no_blending import _MonsterNone
from typing import Dict, List, Tuple, Optional, Unpack, Set  # noqa: F401
from ..positional_mapping import GPM
import itertools
from ..mcs_mapping import IndexMap, ExtendedFMCSMode
from copy import deepcopy
from rdkit import Chem
from ..unmerge_mapper import Unmerge
from ...error import FragmensteinError, DistanceError


class _MonsterExpand(_MonsterNone):
    """
    A variant of no_blend mode, with a focus on expansion
    """

    def by_expansion(self, primary_name: Optional[str] = None, min_mode_index: int = 0) -> Chem.Mol:
        """
        Get the maps. Find the map with the most atoms covered.
        Use that map as the base map for the other maps.
        """
        # -------------- Get the primary hit ----------------------------
        primary_maps: List[Dict[int, int]]
        # primary_name as None chooses one, else the primary name provided is used:
        primary_name, primary_maps = self._get_primary_maps(primary_name)
        # -------------- Get the secondary hits --------------------------------
        # positional_overlap is used by ``_expand_hit_atom_map_by_overlap``
        # which is called by ``_get_unmerge_expansions``
        positional_overlaps: Dict[Tuple[str, str], Dict[int, int]] = self._compute_overlaps()
        if self.throw_on_discard:
            # The two hits do not overlap. This was a decision of the user, surely.
            positional_overlaps: Dict[Tuple[str, str], Dict[int, int]] = {pairing: mapping for pairing, mapping in positional_overlaps.items() if mapping}
            if len(positional_overlaps) == 0 and len(self.hits) > 1:
                # `positional_overlaps` is always empty if there is only one hit!
                # raise DistanceError(hits=self.hits)
                # TODO Add way to make this fatal on request!
                self.journal.warning(f'No positions overlap of the hits')
        unmergers: List[Unmerge] = self._get_unmerge_expansions(primary_name,
                                                                primary_maps,
                                                                positional_overlaps,
                                                                min_mode_index)
        # Nota bene:
        # The custom map is a Dict of hit names to Dict of indices of hit to indices of followup.
        # The Unmerge.map is a Dict of hit names to List of Dict of indices of followup to indices of hit.
        # it is reversed in the Unmerge object.
        if self.throw_on_discard:
            hit_names = [h.GetProp('_Name') for h in self.hits]
            unmergers: List[Unmerge] = [u for u in unmergers if all([h in u.maps and len(u.maps[h]) and len(u.maps[h][0]) for h in hit_names])]
            if len(unmergers) == 0:
                raise DistanceError(hits=self.hits)
        # -------------- Sort the unmergers --------------------------------
        self.positioned_mol, self.mol_options = self._place_unmerger_expansions(unmergers)
        # ---- custom map sanity -----------------------------------
        if sum(map(len, self.custom_map.values())) and not self._check_custom_map(self.positioned_mol):
            self.journal.debug(f'Custom map sanity check failed for best candidate.')
            for mol in self.mol_options:
                if self._check_custom_map(self.positioned_mol):
                    self.positioned_mol = mol
                    break
            else:
                raise FragmensteinError(f'No custom map satisfied by any of the options.')
        return self.positioned_mol

    def _check_custom_map(self, mol: Chem.Mol) -> bool:
        """
        Check that the custom map is satisfied by the molecule.
        """
        originses: List[List[str]] = self.origin_from_mol(mol)
        mapping: Dict[int, int]
        for name, mapping in self.custom_map.items():
            for hit_i, followup_i in mapping.items():
                if followup_i < 0 and \
                        all([(f'{name}.{hit_i}' not in origin) for origins in originses for origin in origins]):
                    pass  # forbidden correctly — absent from all origins
                elif followup_i < 0:  # damnation
                    self.journal.info('Suboptimal fixing: atom is forbidden from matching, '+\
                                      'but is matched indirectly but not constrained.')
                    followup_i = [o for o, origins in enumerate(originses) for origin in origins if f'{name}.{hit_i}' in origin][0]
                    mol.GetAtomWithIdx(followup_i).SetProp('_origin', 'none')
                    mol.GetAtomWithIdx(followup_i).SetBoolProp('_Novel', True)
                elif any([origin == f'{name}.{hit_i}' for origin in originses[followup_i]]):
                    pass  # mapped correctly
                elif hit_i < 0 and name not in originses[followup_i]:
                    pass  # forbidden correctly
                else:
                    self.journal.info(f'Custom map sanity check failed for a combination for hit {name} idx {hit_i} '+\
                                      f'to followup idx {followup_i} — {originses}')
                    return False
        return True

    def _get_primary_maps(self, primary_name: Optional[str] = None) -> Tuple[str, List[Dict[int, int]]]:
        """
        The primary hit is the hit will most in common with the placed molecule.

        :param primary_name:
        :return:
        """
        if primary_name is None:
            # the list is [{hit_atom_idx: template_atom_idx}, ...]
            maps: Dict[str, List[Dict[int, int]]] = self._compute_maps(broad=True)
            # get the largest maps (not the number of maps which would be `len(l)`)
            get_size = lambda l: len(l[0]) if len(l) else 0  # noqa: E731 Guido doesn't like lambda, but I do
            max_size = max(map(get_size, maps.values()))
            # sorted_maps: Dict[str, List[Dict[int, int]]] = dict(sorted(maps.items(),
            #                                                            key=lambda x: get_size(x[1]),
            #                                                            reverse=True))
            biggest_maps = {k: v for k, v in maps.items() if get_size(v) == max_size}
            # choose the first map
            primary_name = list(biggest_maps.items())[0][0]
            primary_maps: List[Dict[int, int]] = biggest_maps[primary_name]
        else:
            primary: Chem.Mol = self.get_hit_by_name(primary_name)
            primary_maps: List[Dict[int, int]] = self._compute_hit_maps(primary, broad=True)
        self.journal.debug(f"Primary hit: {primary_name} with {len(primary_maps)} Primary maps: {primary_maps}")
        return primary_name, primary_maps

    def _get_unmerge_expansions(self,
                                primary_name: str,
                                primary_maps: List[Dict[int, int]],
                                positional_overlaps: Dict[Tuple[str, str], Dict[int, int]],
                                min_mode_index: int) -> List[Unmerge]:
        """
        Calls _perform_unmerge which calls Unmerge.

        :param primary_name: the hit name. Unlike the other methods, this is not optional.
                for example in ``._get_primary_maps(primary_name)`` it can be None.
        :param primary_maps: the maps for the primary hit. This is returned by ``._get_primary_maps(primary_name)``
        :param positional_overlaps: the positional overlaps. see ``_compute_overlaps``.
        :param min_mode_index: the minimum mode index. see ``get_mcs_mappings``, whose default is 0.
        """
        unmergers = []
        # the no_blend mode does the unmerged based on a dict of optional maps,
        # i.e. the maps do not affect each other. Here it is important that they do.
        # hence each primary map is converted into a set of unmerge maps and the best wins.
        # if there is only one hit, then the primary map is the only unmerge map...
        if len(self.hits) == 1:
            return [self._perform_unmerge(maps={primary_name: primary_maps},
                                          n_poisonous=3,
                                          primary_name=primary_name
                                          )]
        # case: multiple hits
        for primary_map in primary_maps:  #: Dict[int, int]
            # iterate over the hit map and expand to all overlapping atoms
            self.journal.debug(f'primary_map: {primary_map}')
            exp_map: Dict[str, Dict[int, int]] = self._expand_hit_atom_map_by_overlap(primary_name,
                                                                                      primary_map,
                                                                                      positional_overlaps,
                                                                                      self.custom_map)
            self.journal.debug(f'initial expanded map (primary + overlaps): {exp_map}')
            exp_maps = {primary_name: [primary_map]}  # only one primary map!
            accounted_for: Set[int] = {i for i in primary_map.values() if i >= 0}
            # get the maps that are not the primary map
            for other in self.hits:
                other_name: str = other.GetProp('_Name')
                if other_name == primary_name:
                    continue
                mappings: List[Dict[int, int]]
                mode: ExtendedFMCSMode
                mappings, mode = self.get_mcs_mappings(other, self.initial_mol, min_mode_index, exp_map)
                # drop any that are redundant with the primary hit
                mappings = [d for d in mappings if len(set(d.values()) - accounted_for) > 0]
                exp_maps[other_name] = mappings
                self.journal.debug(f'candiate expanded maps: {exp_maps} following: {other_name}')
            # {h: f for h, f in .items() if h >= 0 and f >= 0}
            unmergers.append(self._perform_unmerge(maps=exp_maps,
                                                   n_poisonous=3,
                                                   primary_name=primary_name))
        return unmergers

    def _place_unmerger_expansions(self, unmergers: List[Unmerge]) -> Tuple[Chem.Mol, List[Chem.Mol]]:
        scores: List[int] = []
        mol_options = []
        best_mol = None
        for unmerger in unmergers:
            n_off_atoms: int = unmerger.offness(unmerger.combined, unmerger.combined_map)
            scores.append(len(unmerger.combined_map) - 3 * n_off_atoms)
        max_score = max(scores)
        # if they came out equal keep both...
        for score, unmerger in zip(scores, unmergers):
            if score != max_score:
                continue
            positioned_mol, inner_options = self._place_unmerger(unmerger)
            mol_options.extend(inner_options)
            if best_mol:  # it might get sorted again... so it is not important
                mol_options.insert(0, positioned_mol)
            else:
                best_mol = positioned_mol
        return best_mol, mol_options

    def _compute_overlaps(self) -> Dict[Tuple[str, str], Dict[int, int]]:
        positional_overlaps: Dict[Tuple[str, str], Dict[int, int]] = {}
        for mol1, mol2 in itertools.combinations(self.hits, 2):
            mol1_name: str = mol1.GetProp('_Name')
            mol2_name: str = mol2.GetProp('_Name')
            gpm = GPM.get_positional_mapping(mol1, mol2)
            positional_overlaps[(mol1_name, mol2_name)] = gpm
            positional_overlaps[(mol2_name, mol1_name)] = gpm
        return positional_overlaps

    def _expand_hit_atom_map_by_overlap(self,
                                        hit_name: str,
                                        hit_atom_map: Dict[int, int],
                                        positional_overlaps: Dict[Tuple[str, str], Dict[int, int]],
                                        custom_map: Dict[str, Dict[int, int]]) -> Dict[str, Dict[int, int]]:
        """
        Expanded the custom_map by adding all atoms that are covered by the hit_atom_map.

        :param hit_name:
        :param hit_atom_map:
        :param positional_overlaps:
        :param custom_map:
        :param mode:
        :return: custom_map
        """
        expanded: Dict[str, Dict[int, int]] = deepcopy(custom_map)
        expanded[hit_name] = hit_atom_map
        self.fix_custom_map(expanded)
        for hit_atom_idx, template_atom_idx in hit_atom_map.items():
            for other in self.hits:
                other_name: str = other.GetProp('_Name')
                if other_name == hit_name:
                    continue
                # ------------- deal with atoms that overlap --------------------------
                overlaps: Dict[int, int] = positional_overlaps.get((hit_name, other_name), {})
                if hit_atom_idx in overlaps:
                    # ignore the special overrides
                    if template_atom_idx < 0:
                        continue
                    if hit_atom_idx < 0:
                        continue
                    if overlaps[hit_atom_idx] in expanded[other_name]:
                        continue
                    expanded[other_name][overlaps[hit_atom_idx]] = template_atom_idx
                # ------------- deal with atoms that do not overlap ------------------
                elif template_atom_idx in expanded[other_name].values():
                    pass  # there is a mapping already ?!
                else:  # damn the template_atom_idx
                    expanded[other_name][-2 - hit_atom_idx] = template_atom_idx
        return expanded



from rdkit import Chem
from rdkit.Chem import rdFMCS
# TypedDict & Unpack fixed in .legacy:
from typing import Dict, List, Union, Tuple, Optional, TypeVar, Sequence, TypedDict, Unpack  # noqa
from functools import singledispatchmethod  # monkeypatched by .legacy (for Py3.7)
import itertools

from .types import IndexMap, BasicFMCSMode, ExtendedFMCSMode


class VanillaCompareAtoms(rdFMCS.MCSAtomCompare):
    """
    Give than the following does not work, one has to do it in full:

    .. code-block:: python
        :caption: This will raise an error:
        super().__call__(parameters, mol1, atom_idx1, mol2, atom_idx2)

    This class replicates the vanilla functionality
    """

    def __init__(self, comparison: rdFMCS.AtomCompare = rdFMCS.AtomCompare.CompareAnyHeavyAtom):
        """
        Whereas the atomCompare is an enum, this is a callable class.
        But in parameters there is no compareElement booleans etc. only Isotope...
        In https://github.com/rdkit/rdkit/blob/master/Code/GraphMol/FMCS/Wrap/testFMCS.py
        it is clear one needs to make one's own.
        """
        super().__init__()  # noqa the p_obect (python object aka ctypes.py_object) is not used
        self.comparison = comparison

    def __call__(self,  # noqa signature matches... it is just Boost being Boost
                 parameters: rdFMCS.MCSAtomCompareParameters,
                 mol1: Chem.Mol,
                 atom_idx1: int,
                 mol2: Chem.Mol,
                 atom_idx2: int) -> bool:
        a1: Chem.Atom = mol1.GetAtomWithIdx(atom_idx1)
        a2: Chem.Atom = mol2.GetAtomWithIdx(atom_idx2)
        # ------- isotope ------------------------
        if parameters.MatchIsotope and a1.GetIsotope() != a2.GetIsotope():  # noqa
            return False
        if sum([a1.GetAtomicNum() == 0, a2.GetAtomicNum() == 0]) == 1:
            return False
        elif self.comparison == rdFMCS.AtomCompare.CompareIsotopes and a1.GetIsotope() != a2.GetIsotope():  # noqa
            return False
        elif self.comparison == rdFMCS.AtomCompare.CompareElements and a1.GetAtomicNum() != a2.GetAtomicNum():  # noqa
            return False
        elif self.comparison == rdFMCS.AtomCompare.CompareAnyHeavyAtom \
                and (a1.GetAtomicNum() == 1 or a2.GetAtomicNum() == 1):  # noqa
            return False
        elif self.comparison == rdFMCS.AtomCompare.CompareAny:
            pass
        # ------- valence ------------------------
        if parameters.MatchValences and a1.GetTotalValence() != a2.GetTotalValence():  # noqa
            return False
        # ------- chiral ------------------------
        if parameters.MatchChiralTag and not self.CheckAtomChirality(parameters, mol1, atom_idx1, mol2, atom_idx2):
            return False
        # ------- formal ------------------------
        if parameters.MatchFormalCharge and not self.CheckAtomCharge(parameters, mol1, atom_idx1, mol2, atom_idx2):
            return False
        # ------- ring ------------------------
        if parameters.RingMatchesRingOnly and not self.CheckAtomRingMatch(parameters, mol1, atom_idx1, mol2, atom_idx2):
            return False
        # ------- complete ------------------------
        if parameters.CompleteRingsOnly:
            # don't know its intended to be used
            pass
        # ------- distance ------------------------
        if parameters.MaxDistance:  # integer...
            # todo fill!
            pass
        return True


class SpecialCompareAtoms(VanillaCompareAtoms):
    """
    This works like the ``_get_atom_maps`` did prior to Fragmentein version 0.9.
    The mapping as discussed in GitHub issue #23 is in the format

        mapping = { 'hit1': {1:1,2:5} 'hit2': {3:3,4:4,4:6}}

    The hit index is first, followup index is the second.
    The index `-1` for a followup index is the same as not providing the hit index,
    it is described here solely for clarity not for use.

        mapping = { 'hit1': {1:1,2:5, 3:-1} 'hit2': {3:3,4:4,4:6}}

    The index `-2` for a followup index will result in the hit atom index
    not matching any followup index.

        mapping = { 'hit1': {1:1,2:5, 3:-2} 'hit2': {3:3,4:4,4:6}}

    If ``exclusive_mapping`` argument of __init__ is True,
    then if a followup index is present in one hit, but not in a second hit,
    then no atom of the second hit will match that followup atom.
    A negative index for a hit atom index means that no atom in that hit will match the
    corresponding followup index.

        mapping = { 'hit1': {1:1,2:5,-1:3, -2: 7} 'hit2': {3:3,4:4,4:6}}

    However, a positive integer on a different hit overrides it, therefore,
    in the above followup atom 3 cannot be matched to any atom in hit1, but will match
    atom 3 in hit2. Followup atom 7 will not match either.

    .. code-block:: python

        SpecialCompareAtoms(custom_map=mapping, exclusive_mapping=True)
    """

    def __init__(self,
                 comparison: rdFMCS.AtomCompare = rdFMCS.AtomCompare.CompareAnyHeavyAtom,
                 custom_map: Optional[Dict[str, Dict[int, int]]] = None, exclusive_mapping: bool = True):
        super().__init__(comparison=comparison)  # what is p_object?
        self.custom_map = self.fix_custom_map(custom_map)  # custom as in user-defined
        self.banned = self._get_strict_banned() if exclusive_mapping else self._get_lax_banned()

    def _get_strict_banned(self) -> List[int]:
        """
        A list of followup indices that cannot be unmapped
        called if exclusive_mapping is True
        """
        return [foll_idx for mapping in self.custom_map.values() for foll_idx in mapping.values()]

    def _get_lax_flipped_map(self) -> List[int]:
        """
        A list of followup indices that cannot be mapped as per negative hit index
        called if exclusive_mapping is False
        """
        return [foll_idx for mapping in self.custom_map.values() for hit_idx, foll_idx in mapping.items()
                if hit_idx < 0]

    def get_custom(self, hit_mol: Chem.Mol, hit_atom_idx: int) -> int:
        """
        What idx of followup corresponds to the ``hit_atom_idx`` index of ``hit_mol``?
        If nothing, -1 is returned
        """
        name: str = hit_mol.GetProp('_Name')
        return self.custom_map[name].get(hit_atom_idx, -1)

    def __call__(self,
                 parameters: rdFMCS.MCSAtomCompareParameters,
                 hit: Chem.Mol,
                 hit_atom_idx: int,
                 followup: Chem.Mol,
                 followup_atom_idx: int) -> bool:
        if followup.GetProp('_Name') == hit.GetProp('_Name'):
            # self to self mapping
            return super().__call__(parameters, hit, hit_atom_idx, followup, followup_atom_idx)
        if hit.GetProp('_Name') not in self.custom_map:
            # self to self mapping
            return super().__call__(parameters, hit, hit_atom_idx, followup, followup_atom_idx)
        hit_atom = hit.GetAtomWithIdx(hit_atom_idx)
        followup_atom = followup.GetAtomWithIdx(followup_atom_idx)
        symbols = {hit_atom.GetSymbol(), followup_atom.GetSymbol()}
        # ------- Custom -----------------------
        # get the user defined map target -- see docstring for bla bla
        custom: int = self.get_custom(hit, hit_atom_idx)
        if custom == followup_atom_idx:
            # it is the custom map!
            return True
        elif custom != -1:
            # a different index was given
            # or a non -1 negative number
            return False
        elif followup_atom_idx in self.banned:
            # banned is fully filled if exclusive_mapping was true during init
            # otherwise its user provided negatives
            return False
        else:
            # followup index not assigned
            pass
        # ------- Dummy ------------------------
        # dummy atom cannot match non-dummy atom:
        if '*' in symbols and len(symbols) > 1:
            return False
        # ------- protons ------------------------
        # proton cannot match non-proton:
        if 'H' in symbols and len(symbols) > 1:
            return False
        # ------- vanilla ------------------------
        return super().__call__(parameters, hit, hit_atom_idx, followup, followup_atom_idx)

    @singledispatchmethod
    def get_valid_matches(self,
                          parameters: rdFMCS.MCSAtomCompareParameters,
                          common: Chem.Mol,
                          hit: Chem.Mol,
                          followup: Chem.Mol) -> List[Dict[int, int]]:
        """
        Returns a list of possible matches, each a dictionary of hit to follow indices,
        that obey the criteria of the atomic comparison.
        (Formerly this was a IndexAtom)

        This however does not check if all the atoms in custom are present.
        For that, ``Monster._validate_vs_custom`` is called in the method ``Monster.get_mcs_mappings``,
        after calling ``Monster._get_atom_maps``, which calls this method.
        (``Monster.get_mcs_mappings`` tries different matching schema,
        while ``Monster._get_atom_maps`` is for one single scheme).
        The primary reason why this is so, is that there are two tiers of requirements:

        1. The custom map must be present vs
        2. The custom map may be present, but has to be obeyed.

        parameters can be rdFMCS.MCSAtomCompareParameters or rdFMCS.MCSParameters
        """

        matches = []
        # find the common matches and make sure they match each other:
        # GetSubstructMatches is independent of AtomCompare.
        for hit_match, followup_match in itertools.product(hit.GetSubstructMatches(common, uniquify=False),
                                                           followup.GetSubstructMatches(common, uniquify=False)):
            # re `map(int, hit_match)` I do not know under what condition is it not an int...
            # but it was so in previous iteration
            for h, f in zip(map(int, hit_match), map(int, followup_match)):
                if not self(parameters=parameters,
                            hit=hit,
                            followup=followup,
                            hit_atom_idx=h,
                            followup_atom_idx=f):
                    break  # one index does not match. None should match.
            else:
                matches.append(dict(zip(map(int, hit_match), map(int, followup_match))))
        # remove duplicates
        hash_dict = lambda match: hash(tuple(sorted(match.items(), key=lambda i: i[0])))
        unique_matches = {hash_dict(match): match for match in matches}
        return list(unique_matches.values())

    @get_valid_matches.register
    def _(self,
          parameters: rdFMCS.MCSParameters,
          common: Chem.Mol,
          hit: Chem.Mol,
          followup: Chem.Mol) -> List[IndexMap]:
        return self.get_valid_matches(parameters.AtomCompareParameters, common, hit, followup)

    def fix_custom_map(self,
                          custom_map: Dict[str, Union[Sequence[Tuple[int, int]], Dict[int, int]]]) \
            -> Dict[str, Dict[int, int]]:
        """
        This will be deprecated in the future. As Monster.fix_custom_map is better.

        Make sure its Dict[str, Dict[int, int]]

        There is a bit of confusion about the custom map.
        Converts the custom map from dict of lists of 2-element tuples to dict of dicts.
        """
        if custom_map is None:
            # in Monster {h.GetProp('_Name'): {} for h in self.hits}
            return {}
        assert isinstance(custom_map, dict), 'User defined map has to be mol name to Dict[int, int]'
        for name, hit_map in custom_map.items():
            custom_map[name] = dict(hit_map)
        return custom_map
    



__all__ = ['pyrosetta']

from importlib import util, machinery
import sys
from warnings import warn

class AttributeFilledMock:
    """
    This allows:

    .. code-block:: python
        mock = AttributeFilledMock()
        mock.me
        mock.me.again.over()
    """

    warning_msg = 'PyRosetta is not installed! Yet it has been called. '+\
                  'The mock object taking its place does nothing.'+\
                  'This is bound to raise an error'
    __spec__ = machinery.ModuleSpec('pyrosetta', None)
    __signature__ = None

    def __getattr__(self, attr: str):
        return self

    def __call__(self, *args, **kargs):
        warn('This call does nothing as PyRosetta is not installed', category=RuntimeWarning)
        return self

# ======================================================================================================================

if util.find_spec('pyrosetta'):
    import pyrosetta
else:
    warn('PyRosetta is not installed. A mock object is loaded. Any Igor calls will fail.', category=RuntimeWarning)
    pyrosetta = AttributeFilledMock()
    sys.modules['pyrosetta'] = pyrosetta




import unittest

from fragmenstein import Igor
from fragmenstein.mpro import MProVictor


# ======================================================================================================================
# ======================================================================================================================


# ======================================================================================================================

class VictorCombineTests(unittest.TestCase):

    def setUp(self):
        Igor.init_pyrosetta()

    def test_noncovalent(self):
        MProVictor.quick_reanimation = False
        victor = MProVictor.from_hit_codes(hit_codes=['x0305', 'x1249'])
        victor.combine()
        self.assertLess(victor.mrmsd.mrmsd, 1, f'RMSD great that one ({victor.mrmsd.mrmsd})')
        self.assertLess(victor.ddG, -1, f'ddG {victor.ddG}')

    def test_covalent(self):
        MProVictor.quick_reanimation = False
        victor = MProVictor.from_hit_codes(hit_codes=['x0692', 'x0305', 'x1249'])
        victor.combine()
        self.assertLess(victor.mrmsd.mrmsd, 1.2, f'RMSD great that one ({victor.mrmsd.mrmsd})')
        self.assertLess(victor.ddG, -1, f'ddG {victor.ddG}')

    def test_drift_prevention_glitch(self):
        """
        I am not sure why but I had to specify which atoms mapped to which in the drift prevention step
        else it gives ``RuntimeError: No sub-structure match found between the probe and query mol``
        """
        MProVictor.quick_reanimation = False
        victor = MProVictor.from_hit_codes(hit_codes=['x0426', 'x0540'])
        victor.combine()


if __name__ == '__main__':
    unittest.main()



########################################################################################################################

__doc__ = \
    """
    See GitHub documentation
    """
__author__ = "Matteo Ferla. [Github](https://github.com/matteoferla)"
__email__ = "matteo.ferla@gmail.com"
__date__ = "2022 A.D."
__license__ = "MIT"
__citation__ = ""

import functools
import logging
from rdkit import Chem
import numpy as np
from typing import List, Union, Sequence, Tuple, Optional, Dict, Annotated
from rdkit import Chem
from rdkit.Chem import AllChem
import io, time
# nomenclature idea copied from FEGrow:
from openff.toolkit.topology import Molecule as OFFMolecule  # noqa
from openff.toolkit.topology import Topology as OFFTopology  # noqa
from openmmforcefields.generators import SMIRNOFFTemplateGenerator  # noqa
import openmm as mm
import openmm.app as mma
import openmm.unit as mmu
from pathlib import Path
from collections import defaultdict
import time


class Fritz:
    """
    Fritz is a helper class for Victor for use with OpenMM.
    It replaces Igor, the pyrosetta one.

    The two assistants are utterly different and share no attributes.
    """
    journal = logging.getLogger('Fragmenstein')
    forcefield_names = ('amber14-all.xml', 'implicit/gbn2.xml')
    molar_energy_unit = mmu.kilocalorie_per_mole
    integrator = mm.LangevinMiddleIntegrator(300 * mmu.kelvin, 1 / mmu.picosecond, 0.004 * mmu.picoseconds)

    def init_pyrosetta(self):
        """
        This might be called if bad configuration is done?
        """
        self.journal.critical('Fritz does not use PyRosetta... Why was this called?')

    def __init__(self,
                 prepped_mol: Chem.Mol,
                 pdb_block: str,
                 resn: str = 'LIG',
                 resi: str = 1,
                 chain: str = 'X',
                 restraining_atom_indices: Sequence[int] = (),
                 restraint_k: float = 1000.0,
                 mobile_radius: float = 8.0,
                 ):
        tick: float = time.time()
        self.resn: str = resn.strip()
        self.resi: int = int(resi)
        self.chain: str = chain
        # self.prepped_mol is the "initial" mol
        # Igor does not have this as parameterisation is external
        # self.prepped_mol is returned from `preminimized_undummied_mol`
        # so is technically not Victor.Monster.positioned_mol
        self.prepped_mol: Chem.Mol = AllChem.AddHs(prepped_mol, addCoords=True)
        self.correct_pdbinfo(mol=self.prepped_mol, resn=self.resn, resi=self.resi, chain=self.chain)
        # this is apo
        self.pdb_block: str = pdb_block
        self.holo: mma.Modeller = self.plonk(pdb_block,
                                             self.prepped_mol)  # this is Fritz's plonk —Not Victor unlike Igor
        tock: float = time.time()
        self.journal.debug(f'Holo structure made {tock - tick}')
        self.neighboring_res_idxs: List[int] = []  # computed later
        self.forcefield: mma.ForceField = self.create_forcefield()
        self.simulation = self.create_simulation(model=self.holo,
                                                 restraining_atom_indices=restraining_atom_indices,
                                                 restraint_k=restraint_k,
                                                 mobile_radius=mobile_radius)
        self.apo_simulation = self.create_simulation(model=self.pdbblock_to_PDB(self.pdb_block),
                                                     restraining_atom_indices=[],
                                                     restraint_k=0)
        self.ideal_ligand_simulation = self.create_simulation(model=self.rdkit_to_openMM(self.ideal_mol),
                                                              restraining_atom_indices=[],
                                                              restraint_k=0,
                                                              frozen=False)
        # self.shift_ligand(self.unbound_simulation, mm.Vec3(1_000, 0, 0))
        tyck: float = time.time()
        self.journal.debug(f'Simulation created {tyck - tock}')

    @staticmethod
    def correct_pdbinfo(mol: Chem.Mol, resn: str, resi: int, chain: str):
        """
        In Igor, RDKit to Params fixes it.
        In RDKit MolToPDBBlock https://github.com/rdkit/rdkit/blob/32655f5365e363ce13bd6b28e2e9e2544f8680bd/Code/GraphMol/FileParsers/PDBWriter.cpp#L66
        it is hidden.
        """
        missing = []
        dejavu = []
        counter = defaultdict(int)
        # first pass
        for atom in mol.GetAtoms():
            info: Union[None, Chem.AtomPDBResidueInfo] = atom.GetPDBResidueInfo()
            if info is None:
                missing.append(atom)
                continue
            name: str = info.GetName()
            if name.strip() in dejavu:
                # invalid.
                missing.append(atom)
            else:
                dejavu.append(name.strip())
                atom.SetProp('molFileAlias', name)
                info.SetResidueName(resn)
                info.SetResidueNumber(resi)

        def namegen(symbol):
            counter[symbol] += 1
            name = f'{symbol: >2}{counter[symbol]: <2}'
            if name.strip() in dejavu:
                return namegen(symbol)
            dejavu.append(name.strip())
            return name

        for atom in missing:
            symbol = atom.GetSymbol()
            name = namegen(symbol)
            info = Chem.AtomPDBResidueInfo(atomName=name,
                                           residueName=str(resn),
                                           residueNumber=int(resi),
                                           chainId=str(chain))
            atom.SetPDBResidueInfo(info)
            atom.SetProp('molFileAlias', name)

    def plonk(self,
              apo: Union[str, mma.PDBFile, mma.Modeller],
              mol: Optional[Chem.Mol] = None) -> mma.Modeller:
        """
        Plonk the ligand into the apo structure.
        """
        if mol is None:
            mol = self.prepped_mol
        if isinstance(apo, str):  # PDB block
            apo: mma.PDBFile = self.pdbblock_to_PDB(apo)
        # make a copy:
        holo: mma.Modeller = mma.Modeller(apo.topology, apo.positions)
        lig: mma.Modeller = self.rdkit_to_openMM(mol)
        holo.add(lig.topology, lig.positions)  # noqa mmu.Quantity is okay
        return holo

    def rdkit_to_openMM(self, mol: Chem.Mol) -> mma.Modeller:
        # rdkit AssignStereochemistryFrom3D previously applied
        lig = OFFMolecule.from_rdkit(mol, allow_undefined_stereo=True)
        # minor corrections:
        # there is no need to fix via
        # lig_topo._chains[0]._residues[0].name = 'LIG'
        for r_a, m_a in zip(mol.GetAtoms(), lig.atoms):
            assert r_a.GetSymbol() == m_a.symbol, 'Discrepancy'
            r_name = r_a.GetPDBResidueInfo().GetName()
            m_a.name = r_name
        # convert and merge
        lig_topo: mma.Topology = OFFTopology.from_molecules([lig]).to_openmm()
        lig_pos: mmu.Quantity = lig.conformers[0].to_openmm()
        return mma.Modeller(lig_topo, lig_pos)  # noqa mmu.Quantity is okay

    @staticmethod
    def pdbblock_to_PDB(pdb_block: str) -> mma.PDBFile:
        """
        Read a PDB block (string) into a PDBFile
        """
        assert isinstance(pdb_block, str), 'pdb_block must be a string'
        assert len(pdb_block) > 0, 'pdb_block must be a non-empty string'
        iostr = io.StringIO(pdb_block)
        iostr.seek(0)
        pdb = mma.PDBFile(iostr)
        return pdb

    def create_forcefield(self, forcefield_names: Optional[Sequence[str]] = None) -> mma.ForceField:
        """set up forcefield"""
        molecule = OFFMolecule.from_rdkit(self.ideal_mol, allow_undefined_stereo=True)
        smirnoff = SMIRNOFFTemplateGenerator(molecules=molecule)
        if forcefield_names is None:
            forcefield_names = self.forcefield_names
        forcefield = mma.ForceField(*forcefield_names)
        forcefield.registerTemplateGenerator(smirnoff.generator)
        return forcefield

    @functools.cached_property
    def ideal_mol(self):
        ideal = AllChem.AddHs(self.prepped_mol)
        ideal.RemoveAllConformers()
        AllChem.EmbedMolecule(ideal, enforceChirality=True)
        return ideal

    def create_simulation(self,
                          model: Union[mma.PDBFile, mma.Modeller, None] = None,
                          restraint_k: float = 1_000,
                          restraining_atom_indices: Sequence[int] = (),
                          mobile_radius: float = 8.0,
                          frozen: bool = True,
                          ) -> mma.Simulation:
        """
        Creates a simulation object with the ligand harmonically constrained and the distal parts of the protein frozen.
        """
        # deal with model
        if isinstance(model, mma.Modeller):
            pass
        elif isinstance(model, mma.PDBFile):
            model = mma.Modeller(model.topology, model.positions)
        elif isinstance(self.holo, mma.PDBFile):
            model = mma.Modeller(self.holo.topology, self.holo.positions)
        else:
            model = self.holo
        # add Hydrogens
        model.addHydrogens(self.forcefield, pH=7.0)
        # set up system
        system: mm.System = self.forcefield.createSystem(model.topology,
                                                         nonbondedMethod=mma.NoCutoff,
                                                         nonbondedCutoff=1 * mmu.nanometer,
                                                         constraints=mma.HBonds)
        # restrain (harmonic constrain) the ligand
        if restraint_k:
            self.restrain(system, model, k=restraint_k, atom_indices=restraining_atom_indices)
        simulation = mma.Simulation(model.topology, system, self.integrator)
        simulation.context.setPositions(model.positions)
        # freeze the distant parts of the protein
        if frozen and len(self.neighboring_res_idxs) == 0:
            self.journal.debug('Freezing...')
            self.neighboring_res_idxs: List[int] = self.freeze_distal(simulation,
                                                                      lig_resn=self.resn,
                                                                      radius=mobile_radius)
            self.journal.debug(f'Frozen all bar {self.neighboring_res_idxs}')
        elif frozen:
            self.journal.debug('Freezing prespecified')
            self.freeze(simulation=simulation,
                        unfrozen_resn_idxs=self.neighboring_res_idxs)
        else:
            self.journal.debug('No freezing')
        return simulation

    def restrain(self, system: mm.System, pdb: Union[mma.PDBFile, mma.Modeller],
                 k: float = 1_000.0,
                 atom_indices: Sequence[int] = (),
                 **args):
        """
        This needs to be set before the simulation is created. I dont know why.
        """
        positions: mmu.Quantity
        topology: mma.topology.Topology
        topology, positions = self.get_topo_pos(pdb)
        # https://github.com/openmm/openmm-cookbook/blob/main/notebooks/cookbook/Restraining%20Atom%20Positions.ipynb
        restraint = mm.CustomExternalForce('k*periodicdistance(x, y, z, x0, y0, z0)^2')
        # restraint = mm.CustomExternalForce('k*((x-x0)^2+(y-y0)^2+(z-z0)^2)')
        # simulation.system.addForce(restraint)
        system.addForce(restraint)
        restraint.addGlobalParameter('k', k * mmu.kilojoules_per_mole / mmu.nanometer)
        restraint.addPerParticleParameter('x0')
        restraint.addPerParticleParameter('y0')
        restraint.addPerParticleParameter('z0')

        # positions = simulation.context.getState(getPositions=True).getPositions()
        # ## get offset
        # The order seems to not be altered, so an offset works.
        hydrogen = mma.element.Element.getBySymbol('H')
        mm_atom: mma.topology.Atom
        rd_atom: Chem.Atom
        for mm_atom in topology.atoms():
            if mm_atom.residue.name == self.resn:
                offset = mm_atom.index
                break
        else:
            raise ValueError(f'{self.resn} missing')
        for mm_atom in topology.atoms():
            if mm_atom.residue.name != self.resn:
                continue
            elif mm_atom.element == hydrogen:
                continue
            # atom.index is C-style sequential index, atom.id is PDB "index".
            elif atom_indices and mm_atom.index - offset not in atom_indices:
                continue
            rd_atom: Chem.Atom = self.prepped_mol.GetAtomWithIdx(mm_atom.index - offset)
            if rd_atom.HasProp('_x'):
                # original atom positions
                atomic_xyz: mmu.Quantity = mm.Vec3(float(rd_atom.GetDoubleProp('_x')),
                                                   float(rd_atom.GetDoubleProp('_y')),
                                                   float(rd_atom.GetDoubleProp('_z'))) \
                                           * mmu.angstrom
            else:
                # Unlikely... but some hack may be at play. As these are added by `preminimized_undummied_mol`
                self.journal.debug('No _x property. Using position.')
                atomic_xyz: mmu.Quantity = positions[mm_atom.index]
            restraint.addParticle(mm_atom.index, atomic_xyz)
        self.journal.debug(f'N particles restrained: {restraint.getNumParticles()}')

    @staticmethod
    def get_topo_pos(obj) -> Tuple[mma.topology.Topology, mmu.Quantity]:
        """
        Get the topology and positions from a PDBFile, Modeller or Simulation object
        """
        if isinstance(obj, mma.PDBFile) or isinstance(obj, mma.Modeller):
            pdb = obj
            positions = pdb.positions  # noqa mma.PDBFile.positions is a property
            topology = pdb.topology
        elif isinstance(obj, mma.Simulation):
            simulation = obj
            positions: mmu.Quantity = simulation.context.getState(getPositions=True).getPositions()
            topology: mma.topology.Topology = simulation.topology
        else:
            raise ValueError('obj must be of type PDBFile, Modeller or Simulation')
        return topology, positions

    def freeze_distal(self,
                      simulation: mma.Simulation,
                      lig_resn: Optional[str] = None,
                      radius: float = 8.0,
                      **args
                      ) -> List[int]:
        """
        "Constrain" in the crystallographic solving sense.
        This function freezes _everything_ if applied before simulation is created.
        I cannot figure out why.
        """
        if lig_resn is None:
            lig_resn = self.resn
        # constrain...
        positions: mmu.Quantity
        topology: mma.topology.Topology
        topology, positions = self.get_topo_pos(simulation)
        # ## Get ligand centroid
        ligand_residues: List[mma.topology.Residue] = [res for res in topology.residues() if res.name == lig_resn]
        assert len(ligand_residues) == 1, f'There can only be one ligand. N here: {len(ligand_residues)}'
        ligand_residue: mma.topology.Residue = ligand_residues[0]
        ligand_center: mmu.Quantity = self.get_centroid(ligand_residue, positions)
        # ## Find neighbours
        neighbors: List[int] = list({atom.residue.index for atom in topology.atoms() if
                                     self.distance(positions[atom.index], ligand_center, mmu.angstrom) <= radius})
        self.journal.info(f'unfrozen: {len(neighbors)} residues')
        ex_neighbors: List[int] = neighbors + [ligand_residue.index]
        self.freeze(simulation=simulation, unfrozen_resn_idxs=ex_neighbors)
        return ex_neighbors

    def freeze(self, simulation: mma.Simulation, unfrozen_resn_idxs: List[int]):
        """
        Freeze the distal parts of the protein,
        ie. the residues that are not in `unfrozen_resn_idxs`
        This is called by `freeze_distal`.
        """
        positions: mmu.Quantity
        topology: mma.topology.Topology
        topology, positions = self.get_topo_pos(simulation)
        # freezing by setting mass to zero
        n = 0
        for atom in topology.atoms():
            if atom.residue.index not in unfrozen_resn_idxs:
                simulation.system.setParticleMass(atom.index, 0. * mmu.amu)  # Dalton
            else:
                n += 1
        self.journal.info(f'{n} atoms were not frozen')
        return unfrozen_resn_idxs

    @staticmethod
    def get_centroid(residue: mma.topology.Residue, positions: mmu.Quantity) -> mmu.Quantity:
        center: mmu.Quantity = mm.Vec3(0, 0, 0) * mmu.nanometer
        for atom in residue.atoms():
            center += positions[atom.index]
        return center / len(list(residue.atoms()))

    @staticmethod
    def distance(a: mmu.Quantity, b: mmu.Quantity, unit: mmu.Unit) -> np.float64:
        """
        The output is a float, not a Quantity, but multiply by unit to get the Quantity
        """
        d: mm.Vec3 = a.value_in_unit(unit) - b.value_in_unit(unit)
        return np.sqrt(d.x ** 2 + d.y ** 2 + d.z ** 2)

    def shift_ligand(self, simulation: mma.Simulation, amount: Union[mmu.Quantity, mm.Vec3, float]):
        if isinstance(amount, float):
            amount: mmu.Quantity = mm.Vec3(amount, 0, 0) * mmu.angstrom
        elif isinstance(amount, mm.Vec3):
            amount: mmu.Quantity = amount * mmu.angstrom
        else:
            pass
        positions: mmu.Quantity = simulation.context.getState(getPositions=True).getPositions()
        for atom in simulation.topology.atoms():
            if atom.residue.name != self.resn:
                continue
            positions[atom.index] = positions[atom.index] + amount
        simulation.context.setPositions(positions)

    def reanimate(self,
                  tolerance=10 * mmu.kilocalorie_per_mole / (mmu.nano * mmu.meter),
                  maxIterations: int = 0
                  ) -> Dict[str, Union[mmu.Quantity, str]]:
        if isinstance(tolerance, (float, int)):
            tolerance = float(tolerance) * mmu.kilocalorie_per_mole / (mmu.nano * mmu.meter)
        tick: float = time.time()
        self.simulation.minimizeEnergy(tolerance=tolerance, maxIterations=int(maxIterations))
        tock: float = time.time()
        self.journal.debug(f'Reanimation! Minimisation of bound in {tock - tick}s')
        self.apo_simulation.minimizeEnergy(tolerance=tolerance, maxIterations=int(maxIterations))
        self.ideal_ligand_simulation.minimizeEnergy(tolerance=tolerance, maxIterations=int(maxIterations))
        tyck: float = time.time()
        self.journal.debug(f'Reanimation! Minimisation of unbound in {tyck - tock}s')
        # sort data...
        data: Dict[str, Union[mmu.Quantity, str]] = {
            **{f'{k}_bound': v for k, v in self.get_potentials(self.simulation).items()},
            **{f'{k}_apo': v for k, v in self.get_potentials(self.apo_simulation).items()},
            **{f'{k}_ideal': v for k, v in self.get_potentials(self.ideal_ligand_simulation).items()},
            'minimized_pdb': self.to_pdbblock()  # self.simulation by default
        }
        for k in list(data.keys()):
            if '_bound' in k:
                term = k.replace('_bound', '')
                data[f'{term}_unbound'] = data.get(f'{term}_apo', 0) + data.get(f'{term}_ideal', 0)
        data['binding_dG'] = data['total_bound'] - data['total_unbound'] - data.get('CustomExternalForce_bound', 0.)
        return data

    def get_force_by_name(self,
                          name: str,
                          simulation: Optional[mma.Simulation] = None) -> mm.Force:
        if simulation is None:
            simulation = self.simulation
        for f in simulation.system.getForces():
            if f.getName() == name:
                return f
        raise ValueError(f'No force named {name}')

    def get_potentials(self, simulation: Optional[mma.Simulation] = None):
        if simulation is None:
            simulation = self.simulation
        potentials = {}
        potentials['total'] = simulation.context.getState(getEnergy=True).getPotentialEnergy()
        for force in simulation.system.getForces():
            force.setForceGroup(30)
            potentials[force.getName()] = simulation.context \
                .getState(getEnergy=True, groups={30}) \
                .getPotentialEnergy()
            force.setForceGroup(0)
        return potentials

    def remove_potential_by_name(self, name: str, simulation: Optional[mma.Simulation] = None):
        """
        This exists to help do ``.remove_potential_by_name('CustomExternalForce')``
        """
        if simulation is None:
            simulation = self.simulation
        for fi, f in enumerate(simulation.system.getForces()):
            if f.getName() != name:
                continue
            simulation.system.removeForce(fi)
            return True
        return False

    def alter_restraint(self, new_k: float):
        self.journal.warning('`alter_restraint` is experimental and may not work.')
        f: mm.Force = self.get_force_by_name('CustomExternalForce')
        # get the index of the constant named _k_
        for i in range(f.getNumGlobalParameters()):
            if f.getGlobalParameterName(i) == 'k':
                break
        else:
            raise ValueError
        v = f.getGlobalParameterDefaultValue(i)
        f.setGlobalParameterDefaultValue(i, new_k)
        f.updateParametersInContext(self.simulation.context)

    def to_pdbhandle(self, filehandle: io.TextIOWrapper, simulation: Optional[mma.Simulation] = None):
        if simulation is None:
            simulation = self.simulation
        positions: mmu.Quantity = simulation.context.getState(getPositions=True).getPositions()
        mma.PDBFile.writeFile(simulation.topology, positions, filehandle)

    def to_pdbfile(self, filename: str = 'fragmenstein.pdb', simulation: Optional[mma.Simulation] = None):
        with open(filename, 'w') as fh:
            self.to_pdbhandle(fh, simulation)

    def to_pdbblock(self, simulation: Optional[mma.Simulation] = None):
        iostr = io.StringIO()
        self.to_pdbhandle(iostr, simulation)
        iostr.seek(0)
        return iostr.read()

    def to_mol(self, simulation: Optional[mma.Simulation] = None):
        if simulation is None:
            simulation: mma.Simulation = self.simulation
        topo, positions = self.get_topo_pos(simulation)
        lig_vec3: List[mm.Vec3] = [positions[atom.index].value_in_unit(mmu.angstrom) for atom in
                                   simulation.topology.atoms() if atom.residue.name == self.resn]
        n: int = self.prepped_mol.GetNumAtoms()
        if n != len(lig_vec3):
            self.journal.critical('Number of atoms discrepancy!')
        conf = Chem.Conformer(n)
        for mol_i, v in enumerate(lig_vec3):
            conf.SetAtomPosition(mol_i, (v.x, v.y, v.z))
        mol = Chem.Mol(self.prepped_mol)
        mol.RemoveAllConformers()
        mol.AddConformer(conf)
        return mol

    @classmethod
    def minimize_template(cls, apo_block: str,
                          forcefield_names: Optional[Sequence[str]] = None):
        pdb = Fritz.pdbblock_to_PDB(apo_block)
        apo = mma.Modeller(pdb.topology, pdb.positions)
        if forcefield_names is None:
            forcefield_names = cls.forcefield_names
        forcefield = mma.ForceField(*forcefield_names)
        apo.addHydrogens(forcefield, pH=7.0)
        system: mm.System = forcefield.createSystem(apo.topology,
                                                    nonbondedMethod=mma.NoCutoff,
                                                    nonbondedCutoff=1 * mmu.nanometer,
                                                    constraints=mma.HBonds)
        simulation = mma.Simulation(apo.topology, system, cls.integrator)
        simulation.context.setPositions(apo.positions)
        simulation.minimizeEnergy()
        positions: mmu.Quantity = simulation.context.getState(getPositions=True).getPositions()
        iostr = io.StringIO()
        mma.PDBFile.writeFile(simulation.topology, positions, iostr)
        iostr.seek(0)
        return iostr.read()

    def get_unfrozen_pymol_selection(self):
        """
        This is a pymol selection string.
        It acts by residue index only as Chain is being weird:
        `r.chain.id` is `''`, but ought to be `'A'`.
        (Index is 0-based, id is PDB-based w/ gaps and all)
        """
        t, p = self.get_topo_pos(self.simulation)
        _sele = []
        for r in t.residues():
            if r.index in self.neighboring_res_idxs:
                _sele.append(str(r.id))
        return 'resi ' + '+'.join(_sele)



from rdkit import Chem
from ..victor import Victor
from ..m_rmsd import mRMSD
import os, json
from rdkit_to_params import Params
from rdkit import Chem
from rdkit.Chem import AllChem


class Wictor(Victor):
    """
    This Victor does not call Igor

    ... code-block:: python
        from framgenstein.faux_victors import Wictor
        from fragmenstein.demo import Mac1

        Wictor.capture_rdkit_log()
        # Wictor.enable_stdout()
        Wictor.error_to_catch = ()
        wicky = Wictor(hits=Mac1.get_n_filtered_mols(2),
                       pdb_block=Mac1.get_template(),
                       )
        wicky.combine()
    """
    uses_pyrosetta = False

    def _process_settings(self):
        self.journal.debug('Valid settings.py: ff_max_displacement, ff_constraint=1, ff_max_iterations')

    def _calculate_combination_thermo(self):
        # override igor.
        pass

    def _calculate_placement_thermo(self):
        # override igor.
        pass

    def _checkpoint_charlie(self):
        # making folder.
        self.make_output_folder()
        # no igor!
        self._log_warnings()
        self.journal.debug(f'{self.long_name} - saving pose collage')
        min_file = os.path.join(self.work_path, self.long_name, self.long_name + '.holo_minimised.pdb')
        with open(min_file, 'w') as w:
            w.write(self.minimized_pdbblock)
        self.journal.debug(f'{self.long_name} - saving Gibbs')
        # recover bonds
        lig_file = os.path.join(self.work_path, self.long_name, self.long_name + '.minimised.mol')
        Chem.MolToMolFile(self.minimized_mol, lig_file)
        score_file = os.path.join(self.work_path, self.long_name, self.long_name + '.minimised.json')
        with open(score_file, 'w') as w:
            json.dump({'Energy': self.energy_score,
                       'mRMSD': self.mrmsd.mrmsd,
                       'RMSDs': self.mrmsd.rmsds}, w)
        self._log_warnings()

    def post_monster_step(self):
        # this is a black overridable methods that will be the last thing called
        self.params = Params.from_mol(self.monster.positioned_mol, name=self.ligand_resn, generic=False)
        self.params.NAME = self.ligand_resn  # force it.
        self.params.polish_mol()
        self.params.comments.clear()
        self.params.comments.append('Generated via Fragmenstein')
        mol = Chem.Mol(self.monster.positioned_mol)
        if self.settings['ff_use_neighborhood']: # default True
            neighborhood = self.monster.get_neighborhood(self.apo_pdbblock, cutoff=self.settings['ff_neighborhood'])
        else:
            neighborhood = None
        # allow_lax reduces the constraints if it fails
        min_result = self.monster.mmff_minimize(mol,
                                                neighborhood=neighborhood,
                                                ff_max_displacement=float(self.settings['ff_max_displacement']), # def 0
                                                ff_constraint=int(self.settings['ff_constraint']), # def 10
                                                ff_max_iterations=int(self.settings['ff_max_iterations']), # def 200
                                                allow_lax=True)
        self.minimized_mol: Chem.Mol = min_result.mol
        self.minimized_pdbblock: str = self._plonk_monster_in_structure(prepped_mol=self.minimized_mol)
        # The ddG is how strained the molecule is out of the protein... not the drop from binding.
        # recalculating:
        # min_result.ideal is with ff_minimise_ideal True
        ideal: Chem.Mol = self.monster.make_ideal_mol(ff_minimise=bool(self.settings['ff_minimise_ideal']))
        ideal_E: float = ideal.GetDoubleProp('Energy')
        ligand_E: float = self.minimized_mol.GetDoubleProp('Energy')
        # The holo needs recalculating as I don't want the constraints
        AllChem.SanitizeMol(neighborhood)
        hydroneighborhood = Chem.AddHs(neighborhood, addCoords=True)
        holo_E = self.monster.MMFF_score(Chem.CombineMols(self.minimized_mol, hydroneighborhood), delta=False)
        apo_E = self.monster.MMFF_score(hydroneighborhood, delta=False)
        # store data:
        self.energy_score['bound'] = dict(total_score=holo_E, unit='kcal/mol')
        self.energy_score['unbound'] = dict(total_score=apo_E + ideal_E, unit='kcal/mol')
        self.energy_score['apo'] = dict(total_score=apo_E, unit='kcal/mol')
        self.energy_score['ideal'] = dict(total_score=ideal_E, unit='kcal/mol')
        self.energy_score['insitu'] = dict(total_score=ligand_E, unit='kcal/mol')
        self.ddG: float = holo_E - apo_E - ideal_E
        self.mrmsd: mRMSD = self._calculate_rmsd()
        # save to disc
        self._checkpoint_charlie()



"""
For historical reasons the various different modes of placement
were called blending modes —because the hits were blended together.

_MonsterBlend is inherited by MonsterPlace.

This was formerly part of the ``_blend_place.py`` file.
"""

# Coding note. Android-fragments snippets (unrelated to Chem) contaminate the GitHub copilot suggestions.
# Do not trust them.

from ._full import _MonsterFull
from ._partial import _MonsterPartial
# from ._no_blending import _MonsterNone
from ._expand import _MonsterExpand


class _MonsterBlend(_MonsterFull, _MonsterPartial, _MonsterExpand):
    """
    The order of inheritance is:

    1. _MonsterMerge >
    2. _MonsterMap >
    3. _MonsterChimera >
    4. _MonsterRefine >
    5. (_MonsterFull, _MonsterPartial, (_MonsterNone > _MonsterExpand)) >
    6. _MonsterBlend
    """



from ._victor_safety import _VictorSafety
from ._loggerwriter import LoggerWriter
import logging, sys, os, re, requests, unicodedata
from rdkit import Chem
from rdkit.rdBase import WrapLogs, LogToPythonLogger
from rdkit_to_params import Params



class _VictorJournal(_VictorSafety):

    def _log_warnings(self):
        if len(self._warned):
            for w in self._warned:
                self.journal.warning(f'{self.long_name} - {w.message} ({w.category})')
            self._warned.clear()

    @classmethod
    def enable_stdout(cls,
                      level=logging.INFO,
                      captured: bool = True) -> None:
        """
        The ``cls.journal`` is output to the terminal.
        Running it twice can be used to change level.

        :param level: logging level
        :param captured: capture rdkit and pyrosetta?
        :return: None
        """
        cls.journal.handlers = [h for h in cls.journal.handlers if h.name != 'stdout']
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(level)
        handler.set_name('stdout')
        handler.setFormatter(logging.Formatter('[%(asctime)s] %(levelname)s - %(message)s'))
        cls.journal.addHandler(handler)
        # logging.getLogger('py.warnings').addHandler(handler)
        cls.journal.setLevel(level)  # downgrade to the level of the handler
        if captured:
            cls.capture_logs()

    @classmethod
    def enable_logfile(cls, filename='reanimation.log',
                       level=logging.INFO,
                       captured: bool = True) -> None:
        """
        The journal is output to a file.
        Running it twice can be used to change level.

        :param filename: file to write.
        :param level: logging level
        :param captured: capture rdkit and pyrosetta?
        :return: None
        """
        cls.journal.handlers = [h for h in cls.journal.handlers if h.name != 'logfile']
        handler = logging.FileHandler(filename)
        handler.setLevel(level)
        handler.set_name('logfile')
        handler.setFormatter(logging.Formatter('[%(asctime)s] %(levelname)s - %(message)s'))
        cls.journal.addHandler(handler)
        # logging.getLogger('py.warnings').addHandler(handler)
        if captured:
            cls.capture_logs()

    _rdkit_captured = False

    @classmethod
    def capture_rdkit_log(cls):
        """
        RDKit spits a few warning and errors.
        This makes them inline with the logger.
        """
        # formerly:
        #WrapLogs()
        #sys.stderr = LoggerWriter(cls.journal.warning)
        rdkit_log = logging.getLogger('rdkit')
        if not len(cls.journal.handlers):
            rdkit_log.handlers = [logging.NullHandler()]
        else:
            rdkit_log.handlers = cls.journal.handlers
        LogToPythonLogger()
        cls._rdkit_captured = True

    _rosetta_captured = False

    @classmethod
    def capture_rosetta_log(cls):
        """
        Rosetta normally prints to stout. This captures the messages into ``journal``.
        It technically simply passes the handlers of journal to that of the Rosetta logger.
        For alternatives, https://github.com/matteoferla/pyrosetta_scripts/tree/main/init_helper
        """
        if cls._rosetta_captured:
            return
        import pyrosetta
        pyrosetta.logging_support.set_logging_sink()
        logger = logging.getLogger("rosetta")
        logger.setLevel(logging.DEBUG)
        logger.handlers = cls.journal.handlers
        cls._rosetta_captured = True

    @classmethod
    def capture_logs(cls):
        cls.capture_rdkit_log()
        cls.capture_rosetta_log()

    @classmethod
    def slack_me(cls, msg: str) -> bool:
        """
        Send message to a slack webhook

        :param msg: Can be dirty and unicode-y.
        :return: did it work?
        :rtype: bool
        """
        webhook = os.environ['SLACK_WEBHOOK']
        # sanitise.
        msg = unicodedata.normalize('NFKD', msg).encode('ascii', 'ignore').decode('ascii')
        msg = re.sub('[^\w\s\-.,;?!@#()\[\]]', '', msg)
        r = requests.post(url=webhook,
                          headers={'Content-type': 'application/json'},
                          data=f"{{'text': '{msg}'}}")
        if r.status_code == 200 and r.content == b'ok':
            return True
        else:
            return False


# ----------------------- Make params use the same log.

Params.log = _VictorJournal.journal



from typing import Optional, Dict, List, Tuple, Set, Unpack, Union  # noqa cf. .legacy monkeypatch
from warnings import warn

from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem import rdMolAlign

from ._mappings import _MonsterMap

class _MonsterChimera(_MonsterMap):

    def make_chimera(self, template: Chem.Mol, min_mode_index=0) -> Chem.Mol:
        """
        This is to avoid extreme corner corner cases.
        E.g. here the MCS is ringMatchesRingOnly=True and AtomCompare.CompareAny,
        while for the positioning this is not the case.

        Called by full and partial blending modes.

        :return:
        """
        # get the matches
        atom_map, mode = self.get_mcs_mapping(template, self.initial_mol, min_mode_index=min_mode_index)
        follow = {**{k: str(v) for k, v in mode.items()}, 'N_atoms': len(atom_map)}
        self.journal.debug(f"scaffold-followup: {follow}")
        # make the scaffold more like the followup to avoid weird matches.
        chimera = Chem.RWMol(template)
        for scaff_ai, follow_ai in atom_map.items():
            if template.GetAtomWithIdx(scaff_ai).GetSymbol() != self.initial_mol.GetAtomWithIdx(
                    follow_ai).GetSymbol():
                v = {'F': 1, 'Br': 1, 'Cl': 1, 'H': 1, 'B': 3, 'C': 4, 'N': 3, 'O': 2, 'S': 2, 'Se': 2, 'P': 6}
                wanted = self.initial_mol.GetAtomWithIdx(follow_ai)
                if wanted.GetSymbol() == '*':  # all good then!
                    continue
                owned = template.GetAtomWithIdx(scaff_ai)
                diff_valance = owned.GetExplicitValence() - v[wanted.GetSymbol()]
                if wanted.GetSymbol() in ('F', 'Br', 'Cl', 'C', 'H') and diff_valance > 0:
                    continue  # cannot change this.
                elif owned.GetExplicitValence() > 4 and wanted.GetSymbol() not in ('P',):
                    continue
                else:
                    newatom = Chem.Atom(wanted)
                    stdev = chimera.GetAtomWithIdx(scaff_ai).GetDoubleProp('_Stdev')
                    newatom.SetDoubleProp('_Stdev', stdev)
                    origin = chimera.GetAtomWithIdx(scaff_ai).GetProp('_Origin')
                    newatom.SetProp('_Origin', origin)
                    chimera.ReplaceAtom(scaff_ai, newatom)
                    if diff_valance > 0:
                        chimera.GetAtomWithIdx(scaff_ai).SetFormalCharge(diff_valance)
        try:
            chimera.UpdatePropertyCache()  # noqa
        except Chem.AtomValenceException as err:
            warn('Valance issue' + str(err))
        return chimera

    def place_from_map(self,
                       target_mol: Chem.Mol,
                       template_mol: Chem.Mol,
                       atom_map: Optional[Dict] = None,
                       random_seed=None) -> Chem.Mol:
        """
        This method places the atoms with known mapping
        and places the 'uniques' (novel) via an aligned mol (the 'sextant')
        This sextant business is a workaround for the fact that only minimised molecules can use the partial
        embedding function of RDKit.

        The template molecule may be actually two or more fragments,
        as happens for the no blending mode.
        In RDKit, the fragments within a "molecule" are not connected, but have sequential atom indices.

        :param target_mol: target mol
        :param template_mol: the template/scaffold to place the mol
        :param atom_map: something that get_mcs_mapping would return.
        :return:
        """
        # Note none of this malarkey: AllChem.MMFFOptimizeMolecule(ref)
        # prealignment
        if target_mol is None:
            target_mol = self.initial_mol
        sextant = Chem.Mol(target_mol)
        # the target (initial) mol ought to be already sanitised.
        # TODO Check if this is necessary
        # Chem.SanitizeMol(sextant)
        if random_seed:
            kwargs = dict(randomSeed=random_seed)
        else:
            kwargs = {}
        AllChem.EmbedMolecule(sextant, **kwargs)
        AllChem.MMFFOptimizeMolecule(sextant)
        ######################################################
        # mapping retrieval and sextant alignment
        # variables: atom_map sextant -> uniques
        if atom_map is None:
            atom_map, mode = self.get_mcs_mapping(target_mol, template_mol)
            msg = {**{k: str(v) for k, v in mode.items()}, 'N_atoms': len(atom_map)}
            self.journal.debug(f"followup-chimera' = {msg}")
        else:
            self.journal.debug(f'Place from map: using provided atom_map {atom_map}')
        self._add_atom_map_asProp(template_mol, atom_map)
        rdMolAlign.AlignMol(sextant, template_mol, atomMap=list(atom_map.items()), maxIters=500)
        # place atoms that have a known location
        putty = Chem.Mol(sextant)  # this is the molecule returned
        pconf = putty.GetConformer()
        chimera_conf = template_mol.GetConformer()
        uniques = set()  # unique atoms in followup
        for i in range(putty.GetNumAtoms()):
            p_atom = putty.GetAtomWithIdx(i)
            p_atom.SetDoubleProp('_Stdev', 0.)
            p_atom.SetProp('_Origin', 'none')
            if i in atom_map:
                ci = atom_map[i]
                c_atom = template_mol.GetAtomWithIdx(ci)
                if c_atom.HasProp('_Stdev'):
                    stdev = c_atom.GetDoubleProp('_Stdev')
                    origin = c_atom.GetProp('_Origin')
                    p_atom.SetDoubleProp('_Stdev', stdev)
                    p_atom.SetProp('_Origin', origin)
                pconf.SetAtomPosition(i, chimera_conf.GetAtomPosition(ci))
            else:
                uniques.add(i)
        ######################################################
        # I be using a sextant for dead reckoning!
        # variables: sextant unique team
        categories = self._categorize(sextant, uniques)
        done_already = []  # multi-attachment issue.
        for unique_idx in categories['pairs']:  # attachment unique indices
            # check the index was not done already (by virtue of a second attachment)
            if unique_idx in done_already:
                continue
            # get other attachments if any.
            team = self._recruit_team(target_mol, unique_idx, categories['uniques'])
            other_attachments = (team & set(categories['pairs'].keys())) - {unique_idx}
            sights = set()  # atoms to align against
            for att_idx in [unique_idx] + list(other_attachments):
                for pd in categories['pairs'][att_idx]:
                    first_sight = pd['idx']
                    sights.add((first_sight, first_sight))
                    neighs = [i.GetIdx() for i in sextant.GetAtomWithIdx(first_sight).GetNeighbors() if
                              i.GetIdx() not in uniques]
                    for n in neighs:
                        sights.add((n, n))
            if self.attachment and list(categories['dummies']) and list(categories['dummies'])[0] in team:
                r = list(categories['dummies'])[0]
                pconf.SetAtomPosition(r, self.attachment.GetConformer().GetAtomPosition(0))
                sights.add((r, r))
            rdMolAlign.AlignMol(sextant, putty, atomMap=list(sights), maxIters=500)
            sconf = sextant.GetConformer()
            self.journal.debug(f'alignment atoms for {unique_idx} ({team}): {sights}')
            # self.draw_nicely(sextant, highlightAtoms=[a for a, b in sights])
            # copy position over
            for atom_idx in team:
                pconf.SetAtomPosition(atom_idx, sconf.GetAtomPosition(atom_idx))
            # the ring problem does not apply here but would result in rejiggling atoms.

            for other in other_attachments:
                done_already.append(other)
        # complete
        # Why is this performed: TODO Check if this is necessary
        # AllChem.SanitizeMol(putty)
        return putty  # positioned_mol

    def transfer_ring_data(self, donor: Chem.Atom, acceptor: Chem.Atom):
        """
        Transfer the info if a ringcore atom.

        :param donor:
        :param acceptor:
        :return:
        """
        # if donor.GetIntProp('_ori_i') == -1:
        #     data = donor
        pass



# The chemists at Postera were in charge of the data management
# this should probably called something different

import io

import pandas as pd
import requests


def fetch_postera() -> pd.DataFrame:
    """
    Reads the submission file off Github.
    For a local version, just ``postera = read_postera(file)``.
    :return:
    """
    url = "https://raw.githubusercontent.com/postera-ai/" + \
          "COVID_moonshot_submissions/master/covid_submissions_all_info.csv"
    s = requests.get(url).content
    postera = pd.read_csv(io.StringIO(s.decode('utf-8')))
    _add_category(postera)
    return postera


def read_postera(filename: str) -> pd.DataFrame:
    postera = pd.read_csv(filename)
    _add_category(postera)
    return postera


# not exported
def _get_category(row: pd.Series) -> str:
    """
    Postera table has categories as True/False. But it is unlikely that there are multiple.
    Turns out these categories are **not** user submitted.
    However, for consistency with other analysis by other people these are used.

    :param postera: pandas table modified in place
    :return:
    """
    for category in ('Acrylamide', 'Chloroacetamide', 'Vinylsulfonamide', 'Nitrile'):
        if row[category] in ('True', 'true', True):
            return category
    else:
        return 'non-covalent'


def _add_category(postera: pd.DataFrame) -> None:
    postera['category'] = postera.apply(_get_category, axis=1)



import argparse, os, json, itertools, string
from rdkit import Chem
from .._cli_defaults import cli_default_settings
from .base import set_verbose
from typing import List, Any, Dict
from ..laboratory import Laboratory
from ..victor import Victor
import pandas as pd
from rdkit.Chem import PandasTools

class FragmensteinParserPipeline:
    def _define_pipeline(self, parser: argparse.ArgumentParser):
            """fragmenstein pipeline -i hits.sdf -o out.sdf -t template.pdb

            Performs a pipeline run of fragmenstein:
            places the hits against themselves as a reference ("replace"),
            combines the hits,
            searches in SmallWorld for analogues,
            places these,
            and scores the results.
            """
            parser.add_argument('-t', '--template', help='Template PDB file',
                                required=True)
            parser.add_argument('-i', '--input', help='Hits SDF file',
                                required=True)
            parser.add_argument('-o', '--output', help='Output folder',
                                default=cli_default_settings['output'])
            parser.add_argument('-r', '--ranking', help='Ranking method',
                                default=cli_default_settings['ranking'])
            parser.add_argument('-c', '--cutoff', help='Joining cutoff',
                                default=cli_default_settings['cutoff'],
                                type=float)
            parser.add_argument('-q', '--quick',
                                help='Quick reanimation',
                                default=cli_default_settings['quick'],
                                type=bool)
            parser.add_argument('-d', '--sw_dist',
                                help='SmallWorld distance',
                                default=cli_default_settings['sw_dist'],
                                type=int)
            parser.add_argument('-l', '--sw_length', help='SmallWorld length',
                                default=cli_default_settings['sw_length'],
                                type=int)
            parser.add_argument('-b', '--sw_databases', help='SmallWorld databases. Accepts multiple',
                                nargs='+',
                                default=cli_default_settings['sw_databases'])
            parser.add_argument('-s', '--suffix',
                                help='Suffix for output files',
                                default=cli_default_settings['suffix'])
            parser.add_argument('--workfolder',
                                help='Location to put the temp files',
                                default=cli_default_settings['workfolder'])
            parser.add_argument('--victor',
                                help='Which victor to use: Victor, OpenVictor or Wictor',
                                default='Victor')
            parser.add_argument('-n', '--n_cores', help='Number of cores',
                                default=cli_default_settings['n_cores'],
                                type=int)
            parser.add_argument('-m', '--combination_size', help='Number of hits to combine in one step',
                                default=2,
                                type=int)
            parser.add_argument('-k', '--top_mergers', help='Max number of mergers to followup up on',
                                default=cli_default_settings['top_mergers'],
                                type=int)
            parser.add_argument('-e', '--timeout', help='Timeout for each merger',
                                default=cli_default_settings['timeout'],
                                type=int)
            parser.add_argument('-x', '--max_tasks', help='Max number of combinations to try in a batch',
                                default=cli_default_settings['max_tasks'],
                                type=int)
            parser.add_argument('-z', '--blacklist', help='Blacklist file',
                                default=cli_default_settings['blacklist'])
            parser.add_argument('-j', '--weights', help='JSON weights file', default=cli_default_settings['weights'])
            parser.add_argument('-v', '--verbose', action="count", help='verbose')
            parser.set_defaults(func=self.pipeline)
            return parser

    def _pipe_set_up(self, output, cutoff, quick, suffix, **kwargs):
        os.makedirs(output, exist_ok=True)
        Victor.work_path = output
        Victor.monster_throw_on_discard = True  # stop this merger if a fragment cannot be used.
        Victor.monster_joining_cutoff = cutoff  # Å
        Victor.quick_reanimation = quick  # for the impatient
        Victor.error_to_catch = Exception  # stop the whole laboratory otherwise

    def pipeline(self, args: argparse.Namespace):
        """
        Performs a pipeline run of fragmenstein:
        places the hits against themselves as a reference ("replace") to get a baseline score
        combines the hits,
        searches in SmallWorld for analogues,
        places these,
        and scores the results.
        """
        set_verbose(args.verbose)
        settings: Dict[str, Any] = vars(args)
        self._pipe_set_up(**settings)
        with Chem.SDMolSupplier(settings['input'].strip()) as sd:
            # hitdex: Dict[str, Chem.Mol] = {mol.GetProp('_Name'): mol for mol in sd}
            hits: List[Chem.Mol] = list(sd)
        settings['hits'] = hits
        if settings['blacklist']:
            with open(settings['blacklist'].strip()) as fh:
                settings['blacklist'] = [line.strip() for line in fh.readlines()]
        else:
            settings['blacklist'] = cli_default_settings['blacklist']
        print(f'N hits: {len(hits)}')
        with open(settings['template'].strip()) as fh:
            pdbblock = fh.read()
        settings['pdbblock'] = pdbblock
        if settings['weights'] and isinstance(settings['weights'], str):
            with open(settings['weights'].strip()) as fh:
                settings['weights'] = json.load(fh)
        elif settings['weights'] and isinstance(settings['weights'], dict):
            pass  # impossible unless a copy-pasted block.
        else:
            settings['weights'] = cli_default_settings['weights']
        choice = settings.get('victor', 'Victor').lower()
        if choice == 'victor':
            Laboratory.Victor = Victor
        elif choice == 'openvictor':
            from ..openmm.openvictor import OpenVictor
            Laboratory.Victor = OpenVictor
        elif choice == 'wictor':
            from ..faux_victors import Wictor
            Laboratory.Victor = Wictor
        elif choice == 'quicktor':
            from ..faux_victors import Quicktor
            Laboratory.Victor = Quicktor
        else:
            raise ValueError(f'Unknown victor: {choice}')
        Laboratory.Victor.work_path = settings.get('workfolder', cli_default_settings['workfolder'])
        # ## Analyses start here
        # self
        hit_replacements: pd.DataFrame = Laboratory.replace_hits(**settings)
        # run
        max_tasks = settings['max_tasks']
        hitnames = [h.GetProp('_Name') for h in hits]
        all_names = list(map('-'.join, itertools.permutations(hitnames, settings['combination_size'])))
        base_suffix = settings['suffix']
        if max_tasks == 0 or max_tasks > len(all_names):
            all_placements: pd.DataFrame = Laboratory.core_ops(hit_replacements, **settings)
        else:
            all_placements = pd.DataFrame()
            letters = iter(string.ascii_uppercase)
            for i in range(0, len(all_names) + max_tasks, max_tasks):
                settings['suffix'] = base_suffix + next(letters)
                placements: pd.DataFrame = Laboratory.core_ops(hit_replacements, **settings)
                settings['blacklist'] += all_names[i:i + max_tasks]
                all_placements = pd.concat([all_placements, placements], ignore_index=True)
            settings['suffix'] = base_suffix
        Laboratory.correct_weaklings(hit_replacements, all_placements)
        all_placements.to_pickle(f'fragmenstein_placed{base_suffix}.pkl.gz')
        Laboratory.score(all_placements, hit_replacements, **settings)
        all_placements.to_pickle(f'fragmenstein_placed{base_suffix}.pkl.gz')
        all_placements.to_csv(f'fragmenstein_placed{base_suffix}.csv')
        PandasTools.WriteSDF(all_placements, f'fragmenstein_placed{base_suffix}.sdf')
        Laboratory.export_sdf(df=all_placements)



from setuptools import setup, find_packages
from warnings import warn
from importlib import util
import os

this_directory = os.path.abspath(os.path.dirname(__file__))

if os.path.exists(os.path.join(this_directory, 'README.md')):
    with open(os.path.join(this_directory, 'README.md'), 'r') as f:
        long_description = f.read()
else:
    long_description = '''
    Scaffold hopping between bound compounds by stitching them together like a reanimated corpse.
    <img src="https://github.com/matteoferla/Fragmenstein/blob/master/images/fragmenstein.jpg?raw=true" width="300px">

    Documentation in [GitHub](https://github.com/matteoferla/Fragmenstein).

    [![colab demo](https://img.shields.io/badge/Run--demo--in--colab-colab_fragmenstein.ipynb-f9ab00?logo=googlecolab)](https://colab.research.google.com/github/matteoferla/Fragmenstein/blob/master/colab_fragmenstein.ipynb)

    ![Ox](https://upload.wikimedia.org/wikipedia/en/thumb/2/2f/University_of_Oxford.svg/132px-University_of_Oxford.svg.png)
    '''

if os.path.exists(os.path.join(this_directory, 'requirements.txt')):
    with open(os.path.join(this_directory, 'requirements.txt'), 'r') as f:
        requirements = [line.split('#')[0].strip() for line in f.readlines()]
        requirements = [line for line in requirements if line]
else:
    requirements = []

# ---------- Pip and Non pip modules  ----------------------------------------------------------------------------------

  # optional

# `pip install xxx` from wheels (the pre-compiled packages), does not run setup.py.
# While sdist does.
# Therefore these warnings will not be shown for pip installations from wheels...


if not util.find_spec('pyrosetta'):
    warn('The minimisation part of this code uses pyrosetta, which has to be downloaded from ' +
         'the Rosetta software site due to licencing. Without it minimisation via PyRosetta is not possible. '+
         'For openMM usage see OpenVictor.')

if not util.find_spec('pymol2'):
    warn('The module pymol2 is optionally required (conda or apt-get installable).')



setup(
    name='Fragmenstein',
    version='0.13.30',
    description='Merging, linking and placing compounds by stitching them together like a reanimated corpse',
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires='>=3.7',
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    extras_require={'jupyter': ['jupyter']},
    url='https://github.com/matteoferla/Fragmenstein',
    license='MIT',
    author='Matteo Ferla',
    author_email='matteo.ferla@gmail.com',
    classifiers=[ # https://pypi.org/classifiers/
        'Development Status :: 4 - Beta', # Development Status :: 5 - Production/Stable
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    entry_points={
        'console_scripts': ['fragmenstein=fragmenstein.cli:main'],
    }
)



import logging
import tempfile
import unittest, os
# ======================================================================================================================
from multiprocessing import Process

import pyrosetta

# ======================================================================================================================
from rdkit import Chem
from rdkit.Chem import AllChem
from fragmenstein.demo import TestSet
from fragmenstein import Monster, Victor, Igor, mpro_data, Walton, demo
from typing import *
import numpy as np

# ======================================================================================================================

class DummyCombineTests(unittest.TestCase):
    def setUp(self):
        Igor.init_pyrosetta()

    def making_of_dummy(self):
        # This is the code I used to make the dummy.pdb file
        import numpy as np
        import pyrosetta
        from types import ModuleType
        prp: ModuleType = pyrosetta.rosetta.protocols
        prc: ModuleType = pyrosetta.rosetta.core
        pr_scoring: ModuleType = pyrosetta.rosetta.core.scoring
        import pyrosetta_help as ph

        Igor.init_pyrosetta()
        pose = pyrosetta.pose_from_sequence('ELVIS')
        scorefxn: pr_scoring.ScoreFunction = pyrosetta.get_fa_scorefxn()
        prp.relax.FastRelax(scorefxn, 5).apply(pose)
        pose.translate(np.array([-100, 0, 0]))
        block = ph.get_pdbstr(pose)
        return block
    def test_ring_on_line(self):
        """
        This is a case to test what happens to a ring placed on a line:
        It does not open but it does not place with a low RMSD.
        """
        block: str = TestSet.get_text('dummy.pdb')
        hydroaminohexene = Chem.MolFromSmiles('OC=C/C=C/C=CN')
        hydroaminohexene.SetProp('_Name', 'hexane')
        AllChem.EmbedMolecule(hydroaminohexene)
        victor = Victor([hydroaminohexene], pdb_block=block)
        victor.place(Chem.MolFromSmiles('Oc1ccccc1N'))
        self.assertGreater(victor.summarize()['comRMSD'], 1, 'Ring on line did not shift')

if __name__ == '__main__':
    unittest.main()


from IPython.display import SVG, display, HTML
import nglview as nv
from rdkit import Chem
from rdkit.Chem import Draw, AllChem
from ..branding import divergent_colors
from functools import singledispatchmethod
from typing import Tuple, List, Dict, Optional
from io import StringIO
import re
from ..display import MolNGLWidget


def draw(mol, color_map, x=200, y=200, **kwargs):
    from matplotlib.colors import ColorConverter
    d = Draw.rdMolDraw2D.MolDraw2DSVG(x, y)
    d.drawOptions().addAtomIndices = True
    d.drawOptions().addStereoAnnotation = True
    d.drawOptions().prepareMolsBeforeDrawing = False
    d.drawOptions().dummiesAreAttachments = True
    flat = Chem.Mol(mol)
    AllChem.Compute2DCoords(flat)
    color_map2 = {i: ColorConverter().to_rgb(n) for i, n in color_map.items()}
    Draw.rdMolDraw2D.PrepareAndDrawMolecule(d,
                                            flat,
                                            highlightAtoms=list(color_map.keys()),
                                            highlightAtomColors=color_map2,
                                            **kwargs
                                            )
    d.FinishDrawing()
    display(SVG(d.GetDrawingText()))



def get_idx(name, origin, default=None):
    rex = re.match(name + '\.(\d+)', origin)
    if rex:
        return int(rex.group(1))
    return default

class _MonsterUtilCompare:

    def get_color_origins(self) -> Dict[str, Dict[int, str]]:
        """
        Get for the hits and followup the color of the origin
        as seen in show_comparison
        :return:
        """
        color_maps = {}
        origins = self.origin_from_mol(self.positioned_mol)
        n_colors = sum(map(bool, origins))  # bool([]) is sensibly False in Python
        colors = divergent_colors[n_colors]
        mapped_idxs = [i for i, m in enumerate(origins) if m]
        followup_color_map = dict(zip(mapped_idxs, colors))
        positioned_name:str = self.positioned_mol.GetProp('_Name')
        color_maps[positioned_name] = followup_color_map
        # -------------- hits ------------------
        for mol in self.hits:
            name = mol.GetProp('_Name')
            # convert ['MOL.7'] to a list of indices of MOL
            matched = [[get_idx(name, o) for o in ori] for ori in origins]
            # make a color_map
            color_map = {}
            for followup_idx, target_idxs in enumerate(matched):
                for i in target_idxs:  # there ought to only zero or one...
                    if i is None:
                        continue
                    color_map[i] = followup_color_map[followup_idx]
            color_maps[name] = color_map
        return color_maps

    def store_origin_colors_atomically(self):
        """
        Store the color of the origin in the mol as the private property _color
        :return:
        """
        # color_maps = {'benzene': {0: '#AED882', ...},..
        color_maps: Dict[str, Dict[int, str]] = self.get_color_origins()
        for hit in self.hits:  #:Chem.Mol
            name = hit.GetProp('_Name')
            for atom in hit.GetAtoms():  #: Chem.Atom
                atom.SetProp('_color', '#FFFFFF')  # white
            if name not in color_maps:
                continue
            for i in color_maps[name]:
                hit.GetAtomWithIdx(i % 100).SetProp('_color', color_maps[name][i])
        followup_map = color_maps[self.positioned_mol.GetProp('_Name')]
        for atom in self.positioned_mol.GetAtoms():  #: Chem.Atom
            atom.SetProp('_color', '#FFFFFF')  # white
        for i in followup_map:
            self.positioned_mol.GetAtomWithIdx(i).SetProp('_color', followup_map[i])


    def show_comparison(self, *args, **kwargs):
        """"
        Show the atom provenance of the follow molecule.

        Parameters:

        * nothing -> uses self.origin_from_mol
        * hit2followup -> uses the hit2followup dict
        """
        # ------- followup -----------------------
        #  origins is like [['MOL.7'], ['MOL.8'], ['MOL.9'], [], ...
        color_maps: Dict[str, Dict[int, str]] = self.get_color_origins()
        # -------------- hits ------------------
        for mol in self.hits:
            name = mol.GetProp('_Name')
            color_map = color_maps[name]
            print(f'hit {name}')  # legit print, not a debug scar
            draw(mol, color_map, 300, 300, **kwargs)
        print('Followup') # legit print, not a debug scar
        followup_color_map = color_maps[self.positioned_mol.GetProp('_Name')]
        draw(self.positioned_mol, followup_color_map, 300, 300)


    def convert_origins_to_custom_map(self, mol: Optional[Chem.Mol]=None, forbiddance=True) -> Dict[str, Dict[int, int]]:
        """
        The origins stored in the followup differ in format from the custom_map.
        The former is a list of lists of hit_name+atom_index,
        while the latter is a dictionary of hit_name to dictionary of
        hit atom indices to _intended_ followup index.
        This method converts the former to the latter.
        If forbiddance is True, non mapping atoms are marked with negatives.

        If mol is None, then self.positioned_mol is used.

        :return:
        """
        custom_map: Dict[str, Dict[int, int]] = {hit.GetProp('_Name'): {} for hit in self.hits}
        # `origins_from_mol` uses self.positioned_mol by default... what if it is not set?
        origins:List[List[str]] = self.origin_from_mol(mol)  # noqa It is in utility
        for fi, ori in enumerate(origins):
            for origin in ori:
                reg = re.search('(.*)\.(\d+)', str(origin))
                if reg is None:  # literally `none`... Although this is legacy.
                    continue
                name = reg.group(1)
                hi = int(reg.group(2)) % 100
                custom_map[name][hi] = fi
        new_custom_map = {}
        for hit in self.hits:
            name = hit.GetProp('_Name')
            ni = -1
            new_custom_map[name] = {}
            for hi in range(hit.GetNumAtoms()):
                if hi in custom_map[name]:
                    new_custom_map[name][hi] = custom_map[name][hi]
                elif not forbiddance:
                    pass
                elif hit.GetAtomWithIdx(hi).GetAtomicNum() == 1:
                    pass
                else:
                    new_custom_map[name][hi] = ni
                    ni -= 1
        return new_custom_map

    def _to_nglview_and_legend(self, show_positioned_mol:False) -> Tuple[nv.NGLWidget, str]:
        """
        This is called by both ``Monster.to_nglview`` and ``Victor.to_nglview``
        The color can be dictated by the optional private property ``_color``,
        which may have been assigned by ``walton.color_in()`` or the user.

        :return:
        """
        color_series = iter(divergent_colors[len(self.hits)])
        legend = ''
        view = MolNGLWidget()
        for mol in self.hits:
            colorValue = next(color_series) if not mol.HasProp('_color') else mol.GetProp('_color')
            view.add_mol(colorValue=colorValue, mol=mol)
            legend += f'<span style="color: {colorValue}">{mol.GetProp("_Name")}</span> '
        if show_positioned_mol and self.positioned_mol:
            view.add_mol(colorValue='white', mol=self.positioned_mol)
            legend += f'<span>positioned followup (in white)</span> '

        return view, legend

    def to_nglview(self, print_legend: bool = False) -> MolNGLWidget:
        """
        This is not the same method as in Victor.
        generates a NGLWidget (``IPython.display.display`` will show it)
        with the compounds and the merged if present.
        To override the colours:
        The colours will be those in the ``Chem.Mol``'s property ``_color`` if present.

        Returns -> nv.NGLWidget
        """
        view, legend = self._to_nglview_and_legend(show_positioned_mol=True)
        if print_legend:
            display(HTML(legend))
        # async madness: disabled for now.
        # view.center(f'[{self.ligand_resn}]')
        return view

